# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.plugins.gui.messagehooks as messagehooks

class WinTree(messagehooks.MessageHooks):
    """Print Z-Order Desktop Windows Tree"""

    def render_text(self, outfd, data):

        for winsta, atom_tables in data:
            for desktop in winsta.desktops():
                outfd.write("*" * 50 + "\n")
                outfd.write("Window context: {0}\\{1}\\{2}\n\n".format(
                    winsta.dwSessionId, winsta.Name, desktop.Name))
                for wnd, level in desktop.windows(desktop.DeskInfo.spwnd):
                    outfd.write("{0}{1} {2} {3}:{4} {5}\n".format(
                            "." * level,
                            str(wnd.strName or '') or "#{0:x}".format(wnd.head.h),
                            "(visible)" if wnd.Visible else "",
                            wnd.Process.ImageFileName,
                            wnd.Process.UniqueProcessId,
                            self.translate_atom(winsta, atom_tables, wnd.ClassAtom),
                            ))

class Windows(messagehooks.MessageHooks):
    """Print Desktop Windows (verbose details)"""

    def render_text(self, outfd, data):

        for winsta, atom_tables in data:
            for desktop in winsta.desktops():
                outfd.write("*" * 50 + "\n")
                outfd.write("Window context: {0}\\{1}\\{2}\n\n".format(
                    winsta.dwSessionId, winsta.Name, desktop.Name))
                for wnd, _level in desktop.windows(desktop.DeskInfo.spwnd):
                    outfd.write("Window Handle: #{0:x} at {1:#x}, Name: {2}\n".format(
                        wnd.head.h, wnd.obj_offset, str(wnd.strName or '')
                    ))
                    outfd.write("ClassAtom: {0:#x}, Class: {1}\n".format(
                        wnd.ClassAtom,
                        self.translate_atom(winsta, atom_tables, wnd.ClassAtom),
                    ))
                    outfd.write("SuperClassAtom: {0:#x}, SuperClass: {1}\n".format(
                        wnd.SuperClassAtom,
                        self.translate_atom(winsta, atom_tables, wnd.SuperClassAtom),
                    ))
                    outfd.write("pti: {0:#x}, Tid: {1} at {2:#x}\n".format(
                        wnd.head.pti.v(),
                        wnd.Thread.Cid.UniqueThread,
                        wnd.Thread.obj_offset,
                    ))
                    outfd.write("ppi: {0:#x}, Process: {1}, Pid: {2}\n".format(
                        wnd.head.pti.ppi.v(),
                        wnd.Process.ImageFileName,
                        wnd.Process.UniqueProcessId,
                    ))
                    outfd.write("Visible: {0}\n".format("Yes" if wnd.Visible else "No"))
                    outfd.write("Left: {0}, Top: {1}, Bottom: {2}, Right: {3}\n".format(
                        wnd.rcClient.left,
                        wnd.rcClient.top,
                        wnd.rcClient.right, wnd.rcClient.bottom
                    ))
                    outfd.write("Style Flags: {0}\n".format(wnd.style))
                    outfd.write("ExStyle Flags: {0}\n".format(wnd.ExStyle))
                    outfd.write("Window procedure: {0:#x}\n".format(
                        wnd.lpfnWndProc,
                    ))
                    outfd.write("\n")



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

# File-wide pylint filter for protected members, since we have three _BLAH structures
#pylint: disable-msg=W0212

import os
import volatility.utils as utils
import volatility.obj as obj
import volatility.debug as debug
import volatility.win32.tasks as tasks
import volatility.win32.modules as modules
import volatility.plugins.taskmods as taskmods
import volatility.plugins.vadinfo as vadinfo
import volatility.plugins.overlays.windows.windows as windows
import volatility.constants as constants

try:
    import yara
    has_yara = True
except ImportError:
    has_yara = False

try:
    import distorm3
    has_distorm3 = True
except ImportError:
    has_distorm3 = False

#--------------------------------------------------------------------------------
# object classes 
#--------------------------------------------------------------------------------

class MalwareEPROCESS(windows._EPROCESS):
    """Extension of the default EPROCESS with some helpers"""

    @property
    def IsWow64(self):
        """Returns True if this is a wow64 process"""
        return hasattr(self, 'Wow64Process') and self.Wow64Process.v() != 0

    @property
    def SessionId(self):
        """Returns the Session ID of the process"""

        if self.Session.is_valid():
            process_space = self.get_process_address_space()
            if process_space:
                return obj.Object("_MM_SESSION_SPACE",
                                  offset = self.Session,
                                  vm = process_space).SessionId

        return obj.NoneObject("Cannot find process session")

    def get_vads(self, vad_filter = None, skip_max_commit = False):
        """
        Generator for MMVADs that match specific
        metadata.

        @param vad_filter: a callable that is passed the 
        current MMVAD and applies tests to the MMVAD struct
        members or nested struct members. 

        @param skip_max_commit: boolean, if true then VADs
        for Wow64 processes with the MM_MAX_COMMIT flag set
        will not be yielded. 

        @yields a tuple (mmvad, address_space). Where mmvad is
        the MMVAD object in kernel AS and address_space 
        is the process address space. 
        """

        # We absolutely need a process AS. If this 
        # fails then all else fails
        process_space = self.get_process_address_space()
        if not process_space:
            return

        for vad in self.VadRoot.traverse():
            if not vad.is_valid():
                continue
            # Skip Wow64 MM_MAX_COMMIT range
            if (skip_max_commit and self.IsWow64 and vad.u.VadFlags.CommitCharge == 0x7ffffffffffff and 
                    vad.End > 0x7fffffff):
                continue
            # Apply the meta filter if one is supplied
            if vad_filter:
                if not vad_filter(vad):
                    continue
            yield vad, process_space

    def search_process_memory(self, s):
        """
        Search memory for a simple byte string. 
        
        FIXME: as of 2.3 this parameter can also be a list to
        search for mutliple strings concurrently. The 
        single string will be deprecated in 3.0. 

        @param s: the string to search for.

        @returns every occurrance of the string 
        in process memory (as absolute address).
        """

        # Allow for some overlap in case objects are 
        # right on page boundaries 
        overlap = 1024
        
        # Make sure s in a list. This allows you to search for
        # multiple strings at once, without changing the API.
        if type(s) != list:
            debug.warning("Single strings to search_process_memory is deprecated, use a list instead")
            s = [s]

        # All MMVADs that belong to this process.
        for vad, address_space in self.get_vads(skip_max_commit = True):
            offset = vad.Start
            out_of_range = vad.Start + vad.Length
            while offset < out_of_range:
                # Read some data and match it.
                to_read = min(constants.SCAN_BLOCKSIZE + overlap, out_of_range - offset)
                data = address_space.zread(offset, to_read)
                if not data:
                    break
                for x in s:
                    for hit in utils.iterfind(data, x):
                        yield offset + hit
                offset += min(to_read, constants.SCAN_BLOCKSIZE)

    def _injection_filter(self, vad):
        """
        This is a callback that's executed by get_vads()
        when searching for injected code / hidden DLLs. 

        This looks for private allocations that are committed, 
        memory-resident, non-empty (not all zeros) and with an 
        original protection that includes write and execute. 

        It is important to note that protections are applied at 
        the allocation granularity (page level). Thus the original
        protection might not be the current protection, and it
        also might not apply to all pages in the VAD range. 

        @param vad: an MMVAD object.

        @returns: True if the MMVAD looks like it might
        contain injected code. 
        """
        protect = vadinfo.PROTECT_FLAGS.get(vad.u.VadFlags.Protection.v(), "")
        write_exec = "EXECUTE" in protect and "WRITE" in protect

        # The Write/Execute check applies to everything 
        if not write_exec:
            return False

        # This is a typical VirtualAlloc'd injection 
        if vad.u.VadFlags.PrivateMemory == 1 and vad.Tag == "VadS":
            return True

        # This is a stuxnet-style injection 
        if (vad.u.VadFlags.PrivateMemory == 0 and
                protect != "PAGE_EXECUTE_WRITECOPY"):
            return True

        return False

    def _mapped_file_filter(self, vad):
        """
        This is a callback that's executed by get_vads() 
        when searching for memory-mapped files. 

        @param vad: an MMVAD object.

        @returns: True if the MMVAD looks like it might
        contain a mapped file. 
        """

        return vad.u.VadFlags.PrivateMemory == 0 and vad.ControlArea

    def environment_variables(self):
        """Generator for environment variables. 

        The PEB points to our env block - a series of null-terminated
        unicode strings. Each string cannot be more than 0x7FFF chars. 
        End of the list is a quad-null. 
        """

        # Address of the environment block
        if not self.Peb.ProcessParameters.Environment.is_valid():
            return

        process_space = self.get_process_address_space()
        if not process_space:
            return

        block = self.Peb.ProcessParameters.Environment

        s = obj.Object("String", offset = block, vm = process_space,
            encoding = 'utf16', length = 0x7FFF)

        # The terminator is a quad null 
        while len(s):
            if s.count(u"=") == 1:
                yield s.split(u"=")
            # Scan forward the length of this string plus the null
            next_offset = s.obj_offset + ((len(s) + 1) * 2)
            s = obj.Object("String", offset = next_offset,
                vm = process_space, encoding = 'utf16', length = 0x7FFF)

#--------------------------------------------------------------------------------
# profile modifications  
#--------------------------------------------------------------------------------

class MalwareObjectClasesXP(obj.ProfileModification):
    before = ['WindowsObjectClasses']
    conditions = {'os': lambda x: x == 'windows'}
    def modification(self, profile):
        profile.object_classes.update({
            '_EPROCESS': MalwareEPROCESS,
        })

#--------------------------------------------------------------------------------
# functions 
#--------------------------------------------------------------------------------

def Disassemble(data, start, bits = '32bit', stoponret = False):
    """Dissassemble code with distorm3. 

    @param data: python byte str to decode
    @param start: address where `data` is found in memory
    @param bits: use 32bit or 64bit decoding 
    @param stoponret: stop disasm when function end is reached
    
    @returns: tuple of (offset, instruction, hex bytes)
    """

    if not has_distorm3:
        raise StopIteration

    if bits == '32bit':
        mode = distorm3.Decode32Bits
    else:
        mode = distorm3.Decode64Bits

    for o, _, i, h in distorm3.DecodeGenerator(start, data, mode):
        if stoponret and i.startswith("RET"):
            raise StopIteration
        yield o, i, h

#--------------------------------------------------------------------------------
# scanners by scudette
#
# unfortunately the existing scanning framework (i.e. scan.BaseScanner) has 
# some shortcomings that don't allow us to integrate yara easily. 
#
# FIXME: these may need updating after resolving issue 310 which aims to 
# enhance the scan.BaseScanner to better support things like this
#--------------------------------------------------------------------------------

class BaseYaraScanner(object):
    """An address space scanner for Yara signatures."""
    overlap = 1024

    def __init__(self, address_space = None, rules = None):
        self.rules = rules
        self.address_space = address_space

    def scan(self, offset, maxlen):
        # Start scanning from offset until maxlen:
        i = offset

        while i < offset + maxlen:
            # Read some data and match it.
            to_read = min(constants.SCAN_BLOCKSIZE + self.overlap, offset + maxlen - i)
            data = self.address_space.zread(i, to_read)
            if data:
                for match in self.rules.match(data = data):
                    # We currently don't use name or value from the 
                    # yara results but they can be yielded in the 
                    # future if necessary. 
                    for moffset, _name, _value in match.strings:
                        if moffset < constants.SCAN_BLOCKSIZE:
                            yield match, moffset + i

            i += constants.SCAN_BLOCKSIZE

class VadYaraScanner(BaseYaraScanner):
    """A scanner over all memory regions of a process."""

    def __init__(self, task = None, **kwargs):
        """Scan the process address space through the Vads.

        Args:
          task: The _EPROCESS object for this task.
        """
        self.task = task
        BaseYaraScanner.__init__(self, address_space = task.get_process_address_space(), **kwargs)

    def scan(self, offset = 0, maxlen = None):
        for vad, self.address_space in self.task.get_vads(skip_max_commit = True):
            for match in BaseYaraScanner.scan(self, vad.Start, vad.Length):
                yield match

class DiscontigYaraScanner(BaseYaraScanner):
    """A Scanner for Discontiguous scanning."""

    def scan(self, start_offset = 0, maxlen = None):
        contiguous_offset = 0
        total_length = 0
        for (offset, length) in self.address_space.get_available_addresses():
            # Skip ranges before the start_offset
            if self.address_space.address_compare(offset, start_offset) == -1:
                continue

            # Skip ranges that are too high (if maxlen is specified)
            if maxlen != None:
                if self.address_space.address_compare(offset, start_offset + maxlen) > 0:
                    continue

            # Try to join up adjacent pages as much as possible.
            if offset == contiguous_offset + total_length:
                total_length += length
            else:
                # Scan the last contiguous range.
                for match in BaseYaraScanner.scan(self, contiguous_offset, total_length):
                    yield match

                # Reset the contiguous range.
                contiguous_offset = offset
                total_length = length

        if total_length > 0:
            # Do the last range.
            for match in BaseYaraScanner.scan(self, contiguous_offset, total_length):
                yield match

#--------------------------------------------------------------------------------
# yarascan
#--------------------------------------------------------------------------------

class YaraScan(taskmods.DllList):
    "Scan process or kernel memory with Yara signatures"

    def __init__(self, config, *args, **kwargs):
        taskmods.DllList.__init__(self, config, *args, **kwargs)
        config.add_option("KERNEL", short_option = 'K', default = False, action = 'store_true',
                        help = 'Scan kernel modules')
        config.add_option("WIDE", short_option = 'W', default = False, action = 'store_true',
                        help = 'Match wide (unicode) strings')
        config.add_option('YARA-RULES', short_option = 'Y', default = None,
                        help = 'Yara rules (as a string)')
        config.add_option('YARA-FILE', short_option = 'y', default = None,
                        help = 'Yara rules (rules file)')
        config.add_option('DUMP-DIR', short_option = 'D', default = None,
                        help = 'Directory in which to dump the files')

    def _compile_rules(self):
        """Compile the YARA rules from command-line parameters. 
        
        @returns: a YARA object on which you can call 'match'
        
        This function causes the plugin to exit if the YARA 
        rules have syntax errors or are not supplied correctly. 
        """
    
        rules = None
    
        try:
            if self._config.YARA_RULES:
                s = self._config.YARA_RULES
                # Don't wrap hex or regex rules in quotes 
                if s[0] not in ("{", "/"): s = '"' + s + '"'
                # Scan for unicode strings 
                if self._config.WIDE: s += "wide"
                rules = yara.compile(sources = {
                            'n' : 'rule r1 {strings: $a = ' + s + ' condition: $a}'
                            })
            elif self._config.YARA_FILE:
                rules = yara.compile(self._config.YARA_FILE)
            else:
                debug.error("You must specify a string (-Y) or a rules file (-y)")
        except yara.SyntaxError, why:
            debug.error("Cannot compile rules: {0}".format(str(why)))
            
        return rules

    def calculate(self):

        if not has_yara:
            debug.error("Please install Yara from code.google.com/p/yara-project")

        addr_space = utils.load_as(self._config)

        rules = self._compile_rules()

        if self._config.KERNEL:

            # Find KDBG so we know where kernel memory begins. Do not assume
            # the starting range is 0x80000000 because we may be dealing with
            # an image with the /3GB boot switch. 
            kdbg = tasks.get_kdbg(addr_space)

            start = kdbg.MmSystemRangeStart.dereference_as("Pointer")

            # Modules so we can map addresses to owners
            mods = dict((addr_space.address_mask(mod.DllBase), mod)
                        for mod in modules.lsmod(addr_space))
            mod_addrs = sorted(mods.keys())

            # There are multiple views (GUI sessions) of kernel memory.
            # Since we're scanning virtual memory and not physical, 
            # all sessions must be scanned for full coverage. This 
            # really only has a positive effect if the data you're
            # searching for is in GUI memory. 
            sessions = []

            for proc in tasks.pslist(addr_space):
                sid = proc.SessionId
                # Skip sessions we've already seen 
                if sid == None or sid in sessions:
                    continue

                session_space = proc.get_process_address_space()
                if session_space == None:
                    continue

                sessions.append(sid)
                scanner = DiscontigYaraScanner(address_space = session_space,
                                               rules = rules)

                for hit, address in scanner.scan(start_offset = start):
                    module = tasks.find_module(mods, mod_addrs, addr_space.address_mask(address))
                    yield (module, address, hit, session_space.zread(address, 1024))

        else:
            for task in self.filter_tasks(tasks.pslist(addr_space)):
                scanner = VadYaraScanner(task = task, rules = rules)
                for hit, address in scanner.scan():
                    yield (task, address, hit, scanner.address_space.zread(address, 1024))

    def render_text(self, outfd, data):

        if self._config.DUMP_DIR and not os.path.isdir(self._config.DUMP_DIR):
            debug.error(self._config.DUMP_DIR + " is not a directory")

        for o, addr, hit, content in data:
            outfd.write("Rule: {0}\n".format(hit.rule))

            # Find out if the hit is from user or kernel mode 
            if o == None:
                outfd.write("Owner: (Unknown Kernel Memory)\n")
                filename = "kernel.{0:#x}.dmp".format(addr)
            elif o.obj_name == "_EPROCESS":
                outfd.write("Owner: Process {0} Pid {1}\n".format(o.ImageFileName,
                    o.UniqueProcessId))
                filename = "process.{0:#x}.{1:#x}.dmp".format(o.obj_offset, addr)
            else:
                outfd.write("Owner: {0}\n".format(o.BaseDllName))
                filename = "kernel.{0:#x}.{1:#x}.dmp".format(o.obj_offset, addr)

            # Dump the data if --dump-dir was supplied
            if self._config.DUMP_DIR:
                path = os.path.join(self._config.DUMP_DIR, filename)
                fh = open(path, "wb")
                fh.write(content)
                fh.close()

            outfd.write("".join(
                ["{0:#010x}  {1:<48}  {2}\n".format(addr + o, h, ''.join(c))
                for o, h, c in utils.Hexdump(content[0:64])
                ]))

#--------------------------------------------------------------------------------
# malfind
#--------------------------------------------------------------------------------

class Malfind(vadinfo.VADDump):
    "Find hidden and injected code"

    def __init__(self, config, *args, **kwargs):
        vadinfo.VADDump.__init__(self, config, *args, **kwargs)
        config.remove_option("BASE")

    def _is_vad_empty(self, vad, address_space):
        """
        Check if a VAD region is either entirely unavailable 
        due to paging, entirely consiting of zeros, or a 
        combination of the two. This helps ignore false positives
        whose VAD flags match task._injection_filter requirements
        but there's no data and thus not worth reporting it. 

        @param vad: an MMVAD object in kernel AS
        @param address_space: the process address space 
        """
        
        PAGE_SIZE = 0x1000
        all_zero_page = "\x00" * PAGE_SIZE

        offset = 0
        while offset < vad.Length:
            next_addr = vad.Start + offset
            if (address_space.is_valid_address(next_addr) and 
                    address_space.read(next_addr, PAGE_SIZE) != all_zero_page):
                return False
            offset += PAGE_SIZE

        return True

    def render_text(self, outfd, data):

        if not has_distorm3:
            debug.warning("For best results please install distorm3")

        if self._config.DUMP_DIR and not os.path.isdir(self._config.DUMP_DIR):
            debug.error(self._config.DUMP_DIR + " is not a directory")

        for task in data:
            for vad, address_space in task.get_vads(vad_filter = task._injection_filter):

                if self._is_vad_empty(vad, address_space):
                    continue

                content = address_space.zread(vad.Start, 64)

                outfd.write("Process: {0} Pid: {1} Address: {2:#x}\n".format(
                    task.ImageFileName, task.UniqueProcessId, vad.Start))

                outfd.write("Vad Tag: {0} Protection: {1}\n".format(
                    vad.Tag, vadinfo.PROTECT_FLAGS.get(vad.u.VadFlags.Protection.v(), "")))

                outfd.write("Flags: {0}\n".format(str(vad.u.VadFlags)))
                outfd.write("\n")

                outfd.write("{0}\n".format("\n".join(
                    ["{0:#010x}  {1:<48}  {2}".format(vad.Start + o, h, ''.join(c))
                    for o, h, c in utils.Hexdump(content)
                    ])))

                outfd.write("\n")
                outfd.write("\n".join(
                    ["{0:#x} {1:<16} {2}".format(o, h, i)
                    for o, i, h in Disassemble(content, vad.Start)
                    ]))

                # Dump the data if --dump-dir was supplied
                if self._config.DUMP_DIR:

                    filename = os.path.join(self._config.DUMP_DIR,
                        "process.{0:#x}.{1:#x}.dmp".format(
                        task.obj_offset, vad.Start))

                    self.dump_vad(filename, vad, address_space)

                outfd.write("\n\n")

#--------------------------------------------------------------------------------
# ldrmodules 
#--------------------------------------------------------------------------------

class LdrModules(taskmods.DllList):
    "Detect unlinked DLLs"

    def render_text(self, outfd, data):

        self.table_header(outfd,
            [("Pid", "8"),
             ("Process", "20"),
             ("Base", "[addrpad]"),
             ("InLoad", "5"),
             ("InInit", "5"),
             ("InMem", "5"),
             ("MappedPath", "")
            ])

        for task in data:
            # Build a dictionary for all three PEB lists where the
            # keys are base address and module objects are the values
            inloadorder = dict((mod.DllBase.v(), mod)
                                for mod in task.get_load_modules())
            ininitorder = dict((mod.DllBase.v(), mod)
                                for mod in task.get_init_modules())
            inmemorder = dict((mod.DllBase.v(), mod)
                                for mod in task.get_mem_modules())

            # Build a similar dictionary for the mapped files 
            mapped_files = {}
            for vad, address_space in task.get_vads(vad_filter = task._mapped_file_filter):
                # Note this is a lot faster than acquiring the full
                # vad region and then checking the first two bytes. 
                if obj.Object("_IMAGE_DOS_HEADER", offset = vad.Start, vm = address_space).e_magic != 0x5A4D:
                    continue
                mapped_files[int(vad.Start)] = str(vad.FileObject.FileName or '')

            # For each base address with a mapped file, print info on 
            # the other PEB lists to spot discrepancies. 
            for base in mapped_files.keys():
                # Does the base address exist in the PEB DLL lists?
                load_mod = inloadorder.get(base, None)
                init_mod = ininitorder.get(base, None)
                mem_mod = inmemorder.get(base, None)
                # Report if the mapped files are in the PEB lists
                self.table_row(outfd,
                        task.UniqueProcessId,
                        task.ImageFileName,
                        base,
                        str(load_mod != None),
                        str(init_mod != None),
                        str(mem_mod != None),
                        mapped_files[base]
                        )
                # Print the full paths and base names in verbose mode 
                if self._config.verbose:
                    if load_mod:
                        outfd.write("  Load Path: {0} : {1}\n".format(load_mod.FullDllName, load_mod.BaseDllName))
                    if init_mod:
                        outfd.write("  Init Path: {0} : {1}\n".format(init_mod.FullDllName, init_mod.BaseDllName))
                    if mem_mod:
                        outfd.write("  Mem Path:  {0} : {1}\n".format(mem_mod.FullDllName, mem_mod.BaseDllName))



# Volatility
# Copyright (C) 2012-13 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#
# Notwithstanding any rights to use the Software granted by the foregoing,
# if entities or individuals have received a Cease & Desist letter from
# the Volatility Project, the Volatility Foundation, or its copyright holders
# for violating the terms of the GPL version 2, those entities (their employees,
# subcontractors, independent contractors, and affiliates) and / or persons
# are granted no such rights and any use by any one or more of them is
# expressly prohibited, in accordance with Section 4 of the GPL version 2.
# Any rights granted to such entities and / or persons by earlier license
# agreements have been previously terminated as to them.

#pylint: disable-msg=C0111

import os
import re
import math
import volatility.obj as obj
import volatility.utils as utils
import volatility.debug as debug
import volatility.win32.tasks as tasks_mod
import volatility.win32.modules as modules
import volatility.plugins.common as common
import volatility.plugins.taskmods as taskmods
import json

#--------------------------------------------------------------------------------
# Constants
#--------------------------------------------------------------------------------

PAGE_SIZE = 0x1000
PAGE_MASK = PAGE_SIZE - 1
IMAGE_EXT = "img"
DATA_EXT = "dat"
FILEOFFSET_MASK = 0xFFFFFFFFFFFF0000
VACB_BLOCK = 0x40000
VACB_ARRAY = 0x80
VACB_OFFSET_SHIFT = 18
VACB_LEVEL_SHIFT = 7
VACB_SIZE_OF_FIRST_LEVEL = 1 << (VACB_OFFSET_SHIFT + VACB_LEVEL_SHIFT)

class _CONTROL_AREA(obj.CType):

    def extract_ca_file(self, unsafe = False):
        """ Extracts a file from a specified CONTROL_AREA

        Attempts to extract the memory resident pages pertaining to a
        particular CONTROL_AREA object.

        Args:
            control_area: Instance of a CONTROL_AREA object
            unsafe: Relax safety constraints for more data

        Returns:
            mdata: List of pages, (physoffset, fileoffset, size) tuples, that are memory resident
            zpad: List of pages, (offset, size) tuples, that not memory resident

        Raises:
        """

        zpad = []
        mdata = []

        # Depending on the particular address space being used we need to
        # determine if the MMPTE will be either 4 or 8 bytes. The x64
        # and IA32_PAE both use 8 byte PTEs. Whereas, IA32 uses 4 byte
        # PTE entries.
        memory_model = self.obj_vm.profile.metadata.get('memory_model', '32bit')
        pae = self.obj_vm.pae 

        if pae:
            mmpte_size = self.obj_vm.profile.get_obj_size("_MMPTEPA")
        else:
            mmpte_size = self.obj_vm.profile.get_obj_size("_MMPTE")

        # Calculate the size of the _CONTROL_AREA object. It is used to find
        # the correct offset for the SUBSECTION object and the size of the
        # CONTROL_AREA can differ between versions of Windows.
        control_area_size = self.size()

        # The segment is used to describe the physical view of the
        # file. We also use this as a semantic check to see if
        # the processing should continue. If the Segment address
        # is invalid, then we return.
        Segment = self.Segment
        if not Segment.is_valid():
            return mdata, zpad

        # The next semantic check validates that the _SEGMENT object
        # points back to the appropriate _CONTROL_AREA object. If the
        # check is invalid, then we return.
        if (self.obj_offset != Segment.ControlArea):
            return mdata, zpad

        # This is a semantic check added to make sure the Segment.SizeOfSegment value
        # is consistant with the Segment.TotalNumberOfPtes. This occurs fequently
        # when traversing through CONTROL_AREA Objects (~5%), often leading to
        # impossible values. Thus, to be conservative we do not proceed if the
        # Segment does not seem sound.
        if Segment.SizeOfSegment != (Segment.TotalNumberOfPtes * PAGE_SIZE):
            return mdata, zpad

        # The _SUBSECTION object is typically found immediately following
        # the CONTROL_AREA object. For Image Section Objects, the SUBSECTIONS
        # typically correspond with the sections found in the PE. On the otherhand,
        # for Data Section Objects, there is typically only a single valid SUBSECTION.
        subsection_offset = self.obj_offset + control_area_size
        #subsection = obj.Object("_SUBSECTION", subsection_offset, self.kaddr_space)
        subsection = obj.Object("_SUBSECTION", subsection_offset, self.obj_vm)

        # This was another check which was inspired by Ruud's code. It
        # verifies that the first SubsectionBaase (Mmst) never starts
        # at the beginning of a page. The UNSAFE option allows us to
        # ignore this constraint. This was necessary for dumping file data
        # for file objects found with filescan (ie $Mft)
        SubsectionBase = subsection.SubsectionBase
        if (SubsectionBase & PAGE_MASK == 0x0) and not unsafe:
            return mdata, zpad

        # We obtain the Subsections associated with this file
        # by traversing the singly linked list.  Ideally, this
        # list should be null (0) terminated. Upon occasion we
        # we have seen instances where the link pointers are
        # undefined (XXX). If we hit an invalid pointer, the we
        # we exit the traversal.
        while subsection.is_valid() and subsection.v() != 0x0:

            if not subsection:
                break

            # This constraint makes sure that the _SUBSECTION object
            # points back to the associated CONTROL_AREA object. Otherwise,
            # we exit the traversal.
            if (self.obj_offset != subsection.ControlArea):
                break

            # Extract subsection meta-data into local variables
            # this helps with performance and not having to do
            # repetitive lookups.
            PtesInSubsection = subsection.PtesInSubsection
            SubsectionBase = subsection.SubsectionBase
            NextSubsection = subsection.NextSubsection

            # The offset into the file is stored implicitely
            # based on the PTE's location within the Subsection.
            StartingSector = subsection.StartingSector
            SubsectionOffset = StartingSector * 0x200

            # This was another check based on something Ruud
            # had done.  We also so instances where DataSectionObjects
            # would hit a SubsectionBase that was paged aligned
            # and hit strange data. In those instances, the
            # MMPTE SubsectionAddress would not point to the associated
            # Subsection. (XXX)
            if (SubsectionBase & PAGE_MASK == 0x0) and not unsafe:
                break

            ptecount = 0
            while (ptecount < PtesInSubsection):

                pteoffset = SubsectionBase + (mmpte_size * ptecount)
                FileOffset = SubsectionOffset + ptecount * 0x1000

                #  The size of MMPTE changes depending on if it is IA32 (4 bytes)
                #  or IA32_PAE/AMD64 (8 bytes).
                objname = "_MMPTE"
                if pae:
                    objname = "_MMPTEPA"
                mmpte = obj.Object(objname, offset = pteoffset, vm = \
                    subsection.obj_vm)

                if not mmpte:
                    ptecount += 1
                    continue

                # First we check if the entry is valid. If the entry is valid
                # then we get the physical offset. The valid entries are actually
                # handled by the hardware.
                if mmpte.u.Hard.Valid == 0x1:

                    # There are some valid Page Table entries where bit 63
                    # is used to specify if the page is executable. This is
                    # maintained by the processor. If it is not executable,
                    # then the bit is set. Within the Intel documentation,
                    # this is known as the Execute-disable (XD) flag. Regardless,
                    # we will use the get_phys_addr method from the address space
                    # to obtain the physical address.
                    ### Should we check the size of the PAGE? Haven't seen
                    # a hit for LargePage.
                    #if mmpte.u.Hard.LargePage == 0x1:
                    #    print "LargePage"
                    physoffset = mmpte.u.Hard.PageFrameNumber << 12
                    mdata.append([physoffset, FileOffset, PAGE_SIZE])
                    ptecount += 1
                    continue

                elif mmpte.u.Soft.Prototype == 0x1:
                # If the entry is not a valid physical address then
                # we check if it contains a pointer back to the SUBSECTION
                # object. If so, the page is in the backing file and we will
                # need to pad to maintain spacial integrity of the file. This
                # check needs to be performed for looking for the transition flag.
                # The prototype PTEs are initialized as MMPTE_SUBSECTION with the
                # SubsectionAddress.

                # On x86 systems that use 4 byte MMPTE , the MMPTE_SUBSECTION
                # stores an "encoded" version of the SUBSECTION object address.
                # The data is relative to global variable (MmSubsectionBase or
                # MmNonPagedPoolEnd) depending on the WhichPool member of
                # _SUBSECTION. This applies to x86 systems running ntoskrnl.exe.
                # If bit 10 is set then it is prototype/subsection

                    if (memory_model == "32bit") and not pae:
                        SubsectionOffset = \
                          ((mmpte.u.Subsect.SubsectionAddressHigh << 7) |
                          (mmpte.u.Subsect.SubsectionAddressLow << 3))
                        #WhichPool = mmpte.u.Subsect.WhichPool
                        #print "mmpte 0x%x ptecount 0x%x sub-32 0x%x pteoffset 0x%x which 0x%x subdelta 0x%x"%(mmpte.u.Long,ptecount,subsection_offset,pteoffset,WhichPool,SubsectionOffset)
                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1
                        continue

                    if memory_model == "64bit" or pae:
                        SubsectionAddress = mmpte.u.Subsect.SubsectionAddress
                    else:
                        SubsectionAddress = mmpte.u.Long

                    if SubsectionAddress == subsection.obj_offset:
                        # sub proto/prot 4c0 420
                        #print "mmpte 0x%x ptecount 0x%x sub 0x%x offset 0x%x"%(mmpte.u.Long,ptecount,SubsectionAddress,pteoffset)
                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1
                        continue
                    elif (SubsectionAddress == (subsection.obj_offset + 4)):
                        # This was a special case seen on IA32_PAE systems where
                        # the SubsectionAddress pointed to subsection.obj_offset+4
                        # (0x420, 0x460, 0x4a0)

                        #print "mmpte 0x%x ptecount 0x%x sub+4 0x%x offset 0x%x"%(mmpte.u.Long,ptecount,SubsectionAddress,pteoffset)
                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1
                        continue
                    else:
                        #print "mmpte 0x%x ptecount 0x%x sub_unk 0x%x offset 0x%x suboffset 0x%x"%(mmpte.u.Long,ptecount,SubsectionAddress,pteoffset,subsection.obj_offset)
                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1
                        continue

                # Check if the entry is a DemandZero entry.
                elif (mmpte.u.Soft.Transition == 0x0):
                    if ((mmpte.u.Soft.PageFileLow == 0x0) and
                     (mmpte.u.Soft.PageFileHigh == 0x0)):
                        # Example entries include: a0,e0
                        #print "mmpte 0x%x ptecount 0x%x zero offset 0x%x subsec 0x%x"%(mmpte.u.Long,ptecount,pteoffset,subsection.obj_offset)
                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1
                    else:
                        #print "mmpte 0x%x ptecount 0x%x paged offset 0x%x subsec 0x%x file 0x%x offset 0x%x"%(mmpte.u.Long,ptecount,pteoffset,subsection.obj_offset,mmpte.u.Soft.PageFileLow,mmpte.u.Soft.PageFileHigh)

                        zpad.append([FileOffset, PAGE_SIZE])
                        ptecount += 1

                # If the entry is not a valid physical address then
                # we also check to see if it is in transition.
                elif mmpte.u.Trans.Transition == 0x1:
                    physoffset = mmpte.u.Trans.PageFrameNumber << 12
                    #print "mmpte 0x%x ptecount 0x%x transition 0x%x offset 0x%x"%(mmpte.u.Long,ptecount,physoffset,pteoffset)

                    mdata.append([physoffset, FileOffset, PAGE_SIZE])
                    ptecount += 1
                    continue
                else:
                    # This is a catch all for all the other entry types.
                    # sub proto/pro 420,4e0,460,4a0 (x64 +0x28)(x32 +4)
                    # other a0,e0,0, (20,60)
                    # 0x80000000
                    #print "mmpte 0x%x ptecount 0x%x other offset 0x%x subsec 0x%x"%(mmpte.u.Long,ptecount,pteoffset,subsection.obj_offset)
                    zpad.append([FileOffset, PAGE_SIZE])
                    ptecount += 1

	    # Traverse the singly linked list to its next member.
            subsection = NextSubsection

        return (mdata, zpad)

class _SHARED_CACHE_MAP(obj.CType):

    def is_valid(self):
        if not obj.CType.is_valid(self):
            return False

        # Added a semantic check to make sure the data is in a sound state. It's better
        # to catch it early.
        FileSize = self.FileSize.QuadPart
        ValidDataLength = self.ValidDataLength.QuadPart
        SectionSize = self.SectionSize.QuadPart
        #print "SectionSize 0x%x < 0 or FileSize < 0x%x ValidDataLength 0x%x"%(SectionSize,FileSize,ValidDataLength)
        #if SectionSize < 0 or (FileSize < ValidDataLength):
        if SectionSize < 0 or ((FileSize < ValidDataLength) and (ValidDataLength != 0x7fffffffffffffff)):
            return False

        return True

    def process_index_array(self, array_pointer, level, limit, vacbary = None):

        """ Recursively process the sparse multilevel VACB index array

        Args:
            array_pointer:    The address of a possible index array
            shared_cache_map: The associated SHARED_CACHE_MAP object
            level:            The current level
            limit:            The level where we abandon all hope. Ideally this is 7
            vacbary:          An array of collected VACBs

        Returns:
            vacbary:          Collected VACBs 
        """
        if vacbary is None:
            vacbary = []

        if level > limit:
            return []

        # Create an array of VACB entries
        VacbArray = obj.Object("Array", offset = array_pointer, \
            vm = self.obj_vm, count = VACB_ARRAY, \
            targetType = "address", parent = self)

        # Iterate through the entries
        for _i in range(0, VACB_ARRAY):
            # Check if the VACB entry is in use
            if VacbArray[_i] == 0x0:
                continue

            Vacbs = obj.Object("_VACB", offset = int(VacbArray[_i]), vm = self.obj_vm)

            # Check if this is a valid VACB entry by verifying
            # the SharedCacheMap member.
            if Vacbs.SharedCacheMap == self.obj_offset:
                # This is a VACB associated with this cache map
                vacbinfo = self.extract_vacb(Vacbs, VACB_BLOCK)
                if vacbinfo:
                    vacbary.append(vacbinfo)
            else:
                #Process the next level of the multi-level array
                vacbary = self.process_index_array(VacbArray[_i], level + 1, limit, vacbary)
                #vacbary = vacbary + _vacbary
        return vacbary

    def extract_vacb(self, vacbs, size):
        """ Extracts data from a specified VACB

        Attempts to extract the memory resident data from a specified
        VACB.

        Args:
            vacbs: 		The VACB object
            size: 		How much data should be read from the VACB 
            shared_cache_map: 	The associated SHARED_CACHE_MAP object

        Returns:
            vacbinfo:    	Extracted VACB meta-information

        """
        # This is used to collect summary information. We will eventually leverage this
        # when creating the externally exposed APIs.
        vacbinfo = {}

        # Check if the Overlay member of _VACB is resident
        # The Overlay member stores information about the FileOffset
        # and the ActiveCount. This is just another proactive check
        # to make sure the objects are seemingly sound.
        if not vacbs.Overlay:
            return vacbinfo

        # We should add another check to make sure that
        # the SharedCacheMap member of the VACB points back
        # to the corresponding SHARED_CACHE_MAP
        if vacbs.SharedCacheMap != self.v():
            return vacbinfo

        # The FileOffset member of VACB is used to denote the
        # offset within the file where the view begins.  Since all
        # views are 256 KB in size, the bottom 16 bits are used to
        # store the number of references to the view.
        FileOffset = vacbs.Overlay.FileOffset.QuadPart

        if not FileOffset:
            return vacbinfo

        ActiveCount = vacbs.Overlay.ActiveCount
        FileOffset = FileOffset & FILEOFFSET_MASK
        BaseAddress = vacbs.BaseAddress.v()

        vacbinfo['foffset'] = int(FileOffset)
        vacbinfo['acount'] = int(ActiveCount)
        vacbinfo['voffset'] = int(vacbs.obj_offset)
        vacbinfo['baseaddr'] = int(BaseAddress)
        vacbinfo['size'] = int(size)

        return vacbinfo

    def extract_scm_file(self):
        """ Extracts a file from a specified _SHARED_CACHE_MAP

        Attempts to extract the memory resident pages pertaining to a
        particular _SHARED_CACHE_MAP object.

        Args:
            shared_cache_map: Instance of a _SHARED_CACHE_MAP object

        Returns:
            vacbary: List of collected VACB meta information.

        Raises:

        """

        vacbary = []

        if self.obj_offset == 0x0:
            return

        # Added a semantic check to make sure the data is in a sound state.
        #FileSize = shared_cache_map.FileSize.QuadPart
        #ValidDataLength = shared_cache_map.ValidDataLength.QuadPart
        SectionSize = self.SectionSize.QuadPart

        # Let's begin by determining the number of Virtual Address Control
        # Blocks (VACB) that are stored within the cache (nonpaged). A VACB
        # represents one 256-KB view in the system cache. There a are a couple
        # options to use for the data size: ValidDataLength, FileSize,
        # and SectionSize.
        full_blocks = SectionSize / VACB_BLOCK
        left_over = SectionSize % VACB_BLOCK

        # As an optimization, the shared cache map object contains a VACB index
        # array of four entries.  The VACB index arrays are arrays of pointers
        # to VACBs, that track which views of a given file are mapped in the cache.
        # For example, the first entry in the VACB index array refers to the first
        # 256 KB of the file. The InitialVacbs can describe a file up to 1 MB (4xVACB).
        iterval = 0
        while (iterval < full_blocks) and (full_blocks <= 4):
            Vacbs = self.InitialVacbs[iterval]
            vacbinfo = self.extract_vacb(Vacbs, VACB_BLOCK)
            if vacbinfo: vacbary.append(vacbinfo)
            iterval += 1

        # We also have to account for the spill over data
        # that is not found in the full blocks.  The first case to
        # consider is when the spill over is still in InitialVacbs.
        if (left_over > 0) and (full_blocks < 4):
            Vacbs = self.InitialVacbs[iterval]
            vacbinfo = self.extract_vacb(Vacbs, left_over)
            if vacbinfo: vacbary.append(vacbinfo)

        # If the file is larger than 1 MB, a seperate VACB index array
        # needs to be allocated. This is based on how many 256 KB blocks
        # would be required for the size of the file. This newly allocated
        # VACB index array is found through the Vacbs member of
        # SHARED_CACHE_MAP.

        Vacbs = self.Vacbs

        if not Vacbs or (Vacbs.v() == 0):
            return vacbary

        # There are a number of instances where the initial value in
        # InitialVacb will also be the fist entry in Vacbs. Thus we
        # ignore, since it was already processed. It is possible to just
        # process again as the file offset is specified for each VACB.
        if  self.InitialVacbs[0].obj_offset == Vacbs.v():
            return vacbary

        # If the file is less than 32 MB than it can be found in
        # a single level VACB index array.
        size_of_pointer = self.obj_vm.profile.get_obj_size("address")

        if not SectionSize > VACB_SIZE_OF_FIRST_LEVEL:

            ArrayHead = Vacbs.v()
            _i = 0
            for _i in range(0, full_blocks):
                vacb_addr = ArrayHead + (_i * size_of_pointer)
                vacb_entry = obj.Object("address", offset = vacb_addr, vm = Vacbs.obj_vm)

                # If we find a zero entry, then we proceed to the next one.
                # If the entry is zero, then the view is not mapped and we
                # skip. We do not pad because we use the FileOffset to seek
                # to the correct offset in the file.
                if not vacb_entry or (vacb_entry.v() == 0x0):
                    continue
                Vacb = obj.Object("_VACB", offset = vacb_entry.v(), vm = self.obj_vm)
                vacbinfo = self.extract_vacb(Vacb, VACB_BLOCK)
                if vacbinfo:
                    vacbary.append(vacbinfo)
            if left_over > 0:
                vacb_addr = ArrayHead + ((_i + 1) * size_of_pointer)
                vacb_entry = obj.Object("address", offset = vacb_addr, vm = Vacbs.obj_vm)

                if not vacb_entry or (vacb_entry.v() == 0x0):
                    return vacbary

                Vacb = obj.Object("_VACB", offset = vacb_entry.v(), vm = self.obj_vm)
                vacbinfo = self.extract_vacb(Vacb, left_over)
                if vacbinfo:
                    vacbary.append(vacbinfo)
            # The file is less than 32 MB, so we can
            # stop processing.
            return vacbary

        # If we get to this point, then we know that the SectionSize is greator than
        # VACB_SIZE_OF_FIRST_LEVEL (32 MB). Then we have a "sparse multilevel index
        # array where each VACB index array is made up of 128 entries. We no
        # longer assume the data is sequential. (Log2 (32 MB) - 18)/7

        #tree_depth = math.ceil((math.ceil(math.log(file_size, 2)) - 18)/7)
        level_depth = math.ceil(math.log(SectionSize, 2))
        level_depth = (level_depth - VACB_OFFSET_SHIFT) / VACB_LEVEL_SHIFT
        level_depth = math.ceil(level_depth)
        limit_depth = level_depth

        if SectionSize > VACB_SIZE_OF_FIRST_LEVEL:

            # Create an array of 128 entries for the VACB index array
            VacbArray = obj.Object("Array", offset = Vacbs.v(), \
                vm = self.obj_vm, count = VACB_ARRAY, \
                targetType = "address", parent = self)

            # We use a bit of a brute force method. We walk the
            # array and if any entry points to the shared cache map
            # object then we extract it. Otherwise, if it is non-zero
            # we attempt to traverse to the next level.
            for _i in range(0, VACB_ARRAY):
                if VacbArray[_i] == 0x0:
                    continue
                Vacb = obj.Object("_VACB", offset = int(VacbArray[_i]), vm = self.obj_vm)
                if  Vacb.SharedCacheMap == self.obj_offset:
                    vacbinfo = self.extract_vacb(Vacb, VACB_BLOCK)
                    if vacbinfo:
                        vacbary.append(vacbinfo)
                else:
                    # The Index is a pointer
                    #Process the next level of the multi-level array
                    # We set the limit_depth to be the depth of the tree
                    # as determined from the size and we initialize the
                    # current level to 2.
                    vacbary = self.process_index_array(VacbArray[_i], 2, limit_depth, vacbary)
                    #vacbary = vacbary + _vacbary

        return vacbary

class ControlAreaModification(obj.ProfileModification):
    conditions = {'os': lambda x: x == 'windows'}

    def modification(self, profile):
        profile.object_classes.update({
            '_CONTROL_AREA': _CONTROL_AREA,
            '_SHARED_CACHE_MAP': _SHARED_CACHE_MAP,
            })

#--------------------------------------------------------------------------------
# VTypes
#--------------------------------------------------------------------------------

# Windows x86 symbols for ntkrnlpa
ntkrnlpa_types_x86 = {
    '__ntkrnlpa' : [ 0x8, {
    'Long' : [ 0x0, ['unsigned long long']],
    'VolatileLong' : [ 0x0, ['unsigned long long']],
    'Hard' : [ 0x0, ['_MMPTE_HARDWARE_64']],
    'Flush' : [ 0x0, ['_HARDWARE_PTE']],
    'Proto' : [ 0x0, ['_MMPTE_PROTOTYPE']],
    'Soft' : [ 0x0, ['_MMPTE_SOFTWARE_64']],
    'TimeStamp' : [ 0x0, ['_MMPTE_TIMESTAMP']],
    'Trans' : [ 0x0, ['_MMPTE_TRANSITION_64']],
    'Subsect' : [ 0x0, ['_MMPTE_SUBSECTION_64']],
    'List' : [ 0x0, ['_MMPTE_LIST']],
    } ],
    '_MMPTEPA' : [ 0x8, {
    'u' : [ 0x0, ['__ntkrnlpa']],
    } ],
    '_MMPTE_SUBSECTION_64' : [ 0x8, {
    'Valid' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 1, native_type = 'unsigned long long')]],
    'Unused0' : [ 0x0, ['BitField', dict(start_bit = 1, end_bit = 5, native_type = 'unsigned long long')]],
    'Protection' : [ 0x0, ['BitField', dict(start_bit = 5, end_bit = 10, native_type = 'unsigned long long')]],
    'Prototype' : [ 0x0, ['BitField', dict(start_bit = 10, end_bit = 11, native_type = 'unsigned long long')]],
    'Unused1' : [ 0x0, ['BitField', dict(start_bit = 11, end_bit = 32, native_type = 'unsigned long long')]],
    'SubsectionAddress' : [ 0x0, ['BitField', dict(start_bit = 32, end_bit = 64, native_type = 'long long')]],
    } ],
    '_MMPTE_TRANSITION_64' : [ 0x8, {
    'Valid' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 1, native_type = 'unsigned long long')]],
    'Write' : [ 0x0, ['BitField', dict(start_bit = 1, end_bit = 2, native_type = 'unsigned long long')]],
    'Owner' : [ 0x0, ['BitField', dict(start_bit = 2, end_bit = 3, native_type = 'unsigned long long')]],
    'WriteThrough' : [ 0x0, ['BitField', dict(start_bit = 3, end_bit = 4, native_type = 'unsigned long long')]],
    'CacheDisable' : [ 0x0, ['BitField', dict(start_bit = 4, end_bit = 5, native_type = 'unsigned long long')]],
    'Protection' : [ 0x0, ['BitField', dict(start_bit = 5, end_bit = 10, native_type = 'unsigned long long')]],
    'Prototype' : [ 0x0, ['BitField', dict(start_bit = 10, end_bit = 11, native_type = 'unsigned long long')]],
    'Transition' : [ 0x0, ['BitField', dict(start_bit = 11, end_bit = 12, native_type = 'unsigned long long')]],
    'PageFrameNumber' : [ 0x0, ['BitField', dict(start_bit = 12, end_bit = 48, native_type = 'unsigned long long')]],
    'Unused' : [ 0x0, ['BitField', dict(start_bit = 48, end_bit = 64, native_type = 'unsigned long long')]],
    }],
   '_MMPTE_HARDWARE_64' : [ 0x8, {
    'Valid' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 1, native_type = 'unsigned long long')]],
    'Dirty1' : [ 0x0, ['BitField', dict(start_bit = 1, end_bit = 2, native_type = 'unsigned long long')]],
    'Owner' : [ 0x0, ['BitField', dict(start_bit = 2, end_bit = 3, native_type = 'unsigned long long')]],
    'WriteThrough' : [ 0x0, ['BitField', dict(start_bit = 3, end_bit = 4, native_type = 'unsigned long long')]],
    'CacheDisable' : [ 0x0, ['BitField', dict(start_bit = 4, end_bit = 5, native_type = 'unsigned long long')]],
    'Accessed' : [ 0x0, ['BitField', dict(start_bit = 5, end_bit = 6, native_type = 'unsigned long long')]],
    'Dirty' : [ 0x0, ['BitField', dict(start_bit = 6, end_bit = 7, native_type = 'unsigned long long')]],
    'LargePage' : [ 0x0, ['BitField', dict(start_bit = 7, end_bit = 8, native_type = 'unsigned long long')]],
    'Global' : [ 0x0, ['BitField', dict(start_bit = 8, end_bit = 9, native_type = 'unsigned long long')]],
    'CopyOnWrite' : [ 0x0, ['BitField', dict(start_bit = 9, end_bit = 10, native_type = 'unsigned long long')]],
    'Unused' : [ 0x0, ['BitField', dict(start_bit = 10, end_bit = 11, native_type = 'unsigned long long')]],
    'Write' : [ 0x0, ['BitField', dict(start_bit = 11, end_bit = 12, native_type = 'unsigned long long')]],
    'PageFrameNumber' : [ 0x0, ['BitField', dict(start_bit = 12, end_bit = 48, native_type = 'unsigned long long')]],
    'reserved1' : [ 0x0, ['BitField', dict(start_bit = 48, end_bit = 52, native_type = 'unsigned long long')]],
    'SoftwareWsIndex' : [ 0x0, ['BitField', dict(start_bit = 52, end_bit = 63, native_type = 'unsigned long long')]],
    'NoExecute' : [ 0x0, ['BitField', dict(start_bit = 63, end_bit = 64, native_type = 'unsigned long long')]],
    } ],
    '_MMPTE_SOFTWARE_64' : [ 0x8, {
    'Valid' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 1, native_type = 'unsigned long long')]],
    'PageFileLow' : [ 0x0, ['BitField', dict(start_bit = 1, end_bit = 5, native_type = 'unsigned long long')]],
    'Protection' : [ 0x0, ['BitField', dict(start_bit = 5, end_bit = 10, native_type = 'unsigned long long')]],
    'Prototype' : [ 0x0, ['BitField', dict(start_bit = 10, end_bit = 11, native_type = 'unsigned long long')]],
    'Transition' : [ 0x0, ['BitField', dict(start_bit = 11, end_bit = 12, native_type = 'unsigned long long')]],
    'UsedPageTableEntries' : [ 0x0, ['BitField', dict(start_bit = 12, end_bit = 22, native_type = 'unsigned long long')]],
    'InStore' : [ 0x0, ['BitField', dict(start_bit = 22, end_bit = 23, native_type = 'unsigned long long')]],
    'Reserved' : [ 0x0, ['BitField', dict(start_bit = 23, end_bit = 32, native_type = 'unsigned long long')]],
    'PageFileHigh' : [ 0x0, ['BitField', dict(start_bit = 32, end_bit = 64, native_type = 'unsigned long long')]],
    } ],
}

class DumpFilesVTypesx86(obj.ProfileModification):
    """This modification applies the vtypes for all 
    versions of 32bit Windows."""
    before = ['WindowsObjectClasses']
    conditions = {'os': lambda x: x == 'windows',
                  'memory_model': lambda x : x == '32bit'}
    def modification(self, profile):
        profile.vtypes.update(ntkrnlpa_types_x86)

class DumpFiles(common.AbstractWindowsCommand):
    """Extract memory mapped and cached files"""

    def __init__(self, config, *args, **kwargs):
        common.AbstractWindowsCommand.__init__(self, config, *args, **kwargs)

        self.kaddr_space = None
        self.filters = []

        config.add_option('REGEX', short_option = 'r',
                      help = 'Dump files matching REGEX',
                      action = 'store', type = 'string')
        config.add_option('IGNORE-CASE', short_option = 'i',
                      help = 'Ignore case in pattern match',
                      action = 'store_true', default = False)
        config.add_option('OFFSET', short_option = 'o', default = None,
                      help = 'Dump files for Process with physical address OFFSET',
                      action = 'store', type = 'int')
        config.add_option('PHYSOFFSET', short_option = 'Q', default = None,
                      help = 'Dump File Object at physical address PHYSOFFSET',
                      action = 'store', type = 'int')
        config.add_option('DUMP-DIR', short_option = 'D', default = None,
                      cache_invalidator = False,
                      help = 'Directory in which to dump extracted files')
        config.add_option('SUMMARY-FILE', short_option = 'S', default = None,
                      cache_invalidator = False,
                      help = 'File where to store summary information')
        config.add_option('PID', short_option = 'p', default = None,
                      help = 'Operate on these Process IDs (comma-separated)',
                      action = 'store', type = 'str')
        config.add_option('NAME', short_option = 'n',
                      help = 'Include extracted filename in output file path',
                      action = 'store_true', default = False)
        config.add_option('UNSAFE', short_option = 'u',
                      help = 'Relax safety constraints for more data',
                      action = 'store_true', default = False)

        # Possible filters include:
        # SharedCacheMap,DataSectionObject,ImageSectionObject,HandleTable,VAD
        config.add_option("FILTER", short_option = 'F', default = None,
                            help = 'Filters to apply (comma-separated)')

    def filter_tasks(self, tasks):
        """ Reduce the tasks based on the user selectable PIDS parameter.

        Returns a reduced list or the full list if config.PIDS not specified.
        """

        if self._config.PID is None:
            return tasks

        try:
            pidlist = [int(p) for p in self._config.PID.split(',')]
        except ValueError:
            debug.error("Invalid PID {0}".format(self._config.PID))

        return [t for t in tasks if t.UniqueProcessId in pidlist]

    def audited_read_bytes(self, vm, vaddr, length, pad):
        """ This function provides an audited zread capability

        It performs a similar function to zread, in that it will
        pad "invalid" pages.  The main difference is that it allows
        us to collect auditing information about which pages were actually 
        present and which ones were padded. 

        Args:
            vm: The address space to read the data from. 
            vaddr: The virtual address to start reading the data from.
            length: How many bytes to read
            pad: This argument controls if the unavailable bytes are padded.

        Returns:
            ret: Data that was read
            mdata: List of pages that are memory resident
            zpad: List of pages that not memory resident

        Raises:

        """

        zpad = []
        mdata = []

        vaddr, length = int(vaddr), int(length)

        ret = ''

        while length > 0:
            chunk_len = min(length, PAGE_SIZE - (vaddr % PAGE_SIZE))

            buf = vm.read(vaddr, chunk_len)
            if vm.vtop(vaddr) is None:
                zpad.append([vaddr, chunk_len])
                if pad:
                    buf = '\x00' * chunk_len
                else:
                    buf = ''
            else:
                mdata.append([vaddr, chunk_len])

            ret += buf
            vaddr += chunk_len
            length -= chunk_len

        return ret, mdata, zpad

    def calculate(self):
        """ Finds all the requested FILE_OBJECTS  
        
        Traverses the VAD and HandleTable to find all requested
        FILE_OBJECTS

        """
        # Initialize containers for collecting artifacts.
        control_area_list = []
        shared_maps = []
        procfiles = []

        # These lists are used for object collecting files from
        # both the VAD and handle tables
        vadfiles = []
        handlefiles = []

        # Determine which filters the user wants to see
        self.filters = []
        if self._config.FILTER:
            self.filters = self._config.FILTER.split(',')

        # Instantiate the kernel address space
        self.kaddr_space = utils.load_as(self._config)

        # Check to see if the physical address offset was passed for a
        # particular process. Otherwise, use the whole task list.
        if self._config.OFFSET != None:
            tasks_list = [taskmods.DllList.virtual_process_from_physical_offset(
                self.kaddr_space, self._config.OFFSET)]
        else:
            # Filter for the specified processes
            tasks_list = self.filter_tasks(tasks_mod.pslist(self.kaddr_space))

        # If a regex is specified, build it.
        if self._config.REGEX:
            try:
                if self._config.IGNORE_CASE:
                    file_re = re.compile(self._config.REGEX, re.I)
                else:
                    file_re = re.compile(self._config.REGEX)
            except re.error, e:
                debug.error('Error parsing regular expression: {0:s}'.format(e))

        # Check to see if a specific physical address was specified for a
        # FILE_OBJECT. In particular, this is useful for FILE_OBJECTS that
        # are found with filescan that are not associated with a process
        # For example, $Mft.
        if self._config.PHYSOFFSET:
            file_obj = obj.Object("_FILE_OBJECT", self._config.PHYSOFFSET, self.kaddr_space.base, native_vm = self.kaddr_space)
            procfiles.append((None, [file_obj]))
            #return

        # Iterate through the process list and collect all references to
        # FILE_OBJECTS from both the VAD and HandleTable. Each open handle to a file
        # has a corresponding FILE_OBJECT.
        if not self._config.PHYSOFFSET:
            for task in tasks_list:
                pid = task.UniqueProcessId

                # Extract FILE_OBJECTS from the VAD
                if not self.filters or "VAD" in self.filters:
                    for vad in task.VadRoot.traverse():
                        if vad != None:
                            try:
                                control_area = vad.ControlArea
                                if not control_area:
                                    continue
                                file_object = vad.FileObject
                                if file_object:

                                    # Filter for specific FILE_OBJECTS based on user defined
                                    # regular expression. (Performance optimization)
                                    if self._config.REGEX:
                                        name = None
                                        if file_object.FileName:
                                            name = str(file_object.file_name_with_device())
                                        if not name:
                                            continue
                                        if not file_re.search(name):
                                            continue

                                    vadfiles.append(file_object)
                            except AttributeError:
                                pass

                if not self.filters or "HandleTable" in self.filters:
                    # Extract the FILE_OBJECTS from the handle table
                    if task.ObjectTable.HandleTableList:
                        for handle in task.ObjectTable.handles():
                            otype = handle.get_object_type()
                            if otype == "File":
                                file_obj = handle.dereference_as("_FILE_OBJECT")

                                if file_obj:

                                    # Filter for specific FILE_OBJECTS based on user defined
                                    # regular expression. (Performance Optimization)
                                    if self._config.REGEX:
                                        name = None
                                        if file_obj.FileName:
                                            name = str(file_obj.file_name_with_device())
                                        if not name:
                                            continue
                                        if not file_re.search(name):
                                            continue

                                    handlefiles.append(file_obj)

                # Append the lists of file objects
                #allfiles = handlefiles + vadfiles
                procfiles.append((pid, handlefiles + vadfiles))

        for pid, allfiles in procfiles:
            for file_obj in allfiles:

                if not self._config.PHYSOFFSET:
                    offset = file_obj.obj_offset
                else:
                    offset = self._config.PHYSOFFSET

                name = None

                if file_obj.FileName:
                    name = str(file_obj.file_name_with_device())

                # The SECTION_OBJECT_POINTERS structure is used by the memory
                # manager and cache manager to store file-mapping and cache information
                # for a particular file stream. We will use it to determine what type
                # of FILE_OBJECT we have and how it should be parsed.
                if file_obj.SectionObjectPointer:
                    DataSectionObject = \
                        file_obj.SectionObjectPointer.DataSectionObject
                    SharedCacheMap = \
                        file_obj.SectionObjectPointer.SharedCacheMap
                    ImageSectionObject = \
                        file_obj.SectionObjectPointer.ImageSectionObject

                    # The ImageSectionObject is used to track state information for
                    # an executable file stream. We will use it to extract memory
                    # mapped binaries.

                    if not self.filters or "ImageSectionObject" in self.filters:

                        if ImageSectionObject and ImageSectionObject != 0:
                            summaryinfo = {}
                            # It points to a image section object( CONTROL_AREA )
                            control_area = \
                                ImageSectionObject.dereference_as('_CONTROL_AREA')

                            if not control_area in control_area_list:
                                control_area_list.append(control_area)

                                # The format of the filenames: file.<pid>.<control_area>.[img|dat]
                                ca_offset_string = "0x{0:x}".format(control_area.obj_offset)
                                if self._config.NAME and name != None:
                                    fname = name.split("\\")
                                    ca_offset_string += "." + fname[-1]
                                file_string = ".".join(["file", str(pid), ca_offset_string, IMAGE_EXT])
                                of_path = os.path.join(self._config.DUMP_DIR, file_string)
                                (mdata, zpad) = control_area.extract_ca_file(self._config.UNSAFE)
                                summaryinfo['name'] = name
                                summaryinfo['type'] = "ImageSectionObject"
                                if pid:
                                    summaryinfo['pid'] = int(pid)
                                else:
                                    summaryinfo['pid'] = None
                                summaryinfo['present'] = mdata
                                summaryinfo['pad'] = zpad
                                summaryinfo['fobj'] = int(offset)
                                summaryinfo['ofpath'] = of_path
                                yield summaryinfo

                    # The DataSectionObject is used to track state information for
                    # a data file stream. We will use it to extract artifacts of
                    # memory mapped data files.

                    if not self.filters or "DataSectionObject" in self.filters:

                        if DataSectionObject and DataSectionObject != 0:
                            summaryinfo = {}
                            # It points to a data section object (CONTROL_AREA)
                            control_area = DataSectionObject.dereference_as('_CONTROL_AREA')

                            if not control_area in control_area_list:
                                control_area_list.append(control_area)

                                # The format of the filenames: file.<pid>.<control_area>.[img|dat]
                                ca_offset_string = "0x{0:x}".format(control_area.obj_offset)
                                if self._config.NAME and name != None:
                                    fname = name.split("\\")
                                    ca_offset_string += "." + fname[-1]
                                file_string = ".".join(["file", str(pid), ca_offset_string, DATA_EXT])
                                of_path = os.path.join(self._config.DUMP_DIR, file_string)

                                (mdata, zpad) = control_area.extract_ca_file(self._config.UNSAFE)
                                summaryinfo['name'] = name
                                summaryinfo['type'] = "DataSectionObject"
                                if pid:
                                    summaryinfo['pid'] = int(pid)
                                else:
                                    summaryinfo['pid'] = None
                                summaryinfo['present'] = mdata
                                summaryinfo['pad'] = zpad
                                summaryinfo['fobj'] = int(offset)
                                summaryinfo['ofpath'] = of_path
                                yield summaryinfo

                    # The SharedCacheMap is used to track views that are mapped to the
                    # data file stream. Each cached file has a single SHARED_CACHE_MAP object,
                    # which has pointers to slots in the system cache which contain views of the file.
                    # The shared cache map is used to describe the state of the cached file.
                    if self.filters and "SharedCacheMap" not in self.filters:
                        continue

                    if SharedCacheMap:
                        vacbary = []
                        summaryinfo = {}
                        #The SharedCacheMap member points to a SHARED_CACHE_MAP object.
                        shared_cache_map = SharedCacheMap.dereference_as('_SHARED_CACHE_MAP')
                        if shared_cache_map.obj_offset == 0x0:
                            continue

                        # Added a semantic check to make sure the data is in a sound state. It's better
                        # to catch it early.
                        if not shared_cache_map.is_valid():
                            continue

                        if not shared_cache_map.obj_offset in shared_maps:
                            shared_maps.append(shared_cache_map.obj_offset)
                        else:
                            continue

                        shared_cache_map_string = ".0x{0:x}".format(shared_cache_map.obj_offset)
                        if self._config.NAME and name != None:
                            fname = name.split("\\")
                            shared_cache_map_string = shared_cache_map_string + "." + fname[-1]
                        of_path = os.path.join(self._config.DUMP_DIR, "file." + str(pid) + shared_cache_map_string + ".vacb")

                        vacbary = shared_cache_map.extract_scm_file()

                        summaryinfo['name'] = name
                        summaryinfo['type'] = "SharedCacheMap"
                        if pid:
                            summaryinfo['pid'] = int(pid)
                        else:
                            summaryinfo['pid'] = None
                        summaryinfo['fobj'] = int(offset)
                        summaryinfo['ofpath'] = of_path
                        summaryinfo['vacbary'] = vacbary
                        yield summaryinfo

    def render_text(self, outfd, data):
        """Renders output for the dumpfiles plugin. 

        This includes extracting the file artifacts from memory 
        to the specified dump directory.

        Args:
            outfd: The file descriptor to write the text to.
            data:  (summaryinfo)

        """

        # Summary file object
        summaryfo = None
        summaryinfo = data

        if self._config.DUMP_DIR == None:
            debug.error("Please specify a dump directory (--dump-dir)")
        if not os.path.isdir(self._config.DUMP_DIR):
            debug.error(self._config.DUMP_DIR + " is not a directory")

        if self._config.SUMMARY_FILE:
            summaryfo = open(self._config.SUMMARY_FILE, 'wb')

        for summaryinfo in data:

            if summaryinfo['type'] == "DataSectionObject":

                outfd.write("DataSectionObject {0:#010x}   {1:<6} {2}\n".format(summaryinfo['fobj'], summaryinfo['pid'], summaryinfo['name']))
                if len(summaryinfo['present']) == 0:
                    continue

                of = open(summaryinfo['ofpath'], 'wb')

                for mdata in summaryinfo['present']:
                    rdata = None
                    if not mdata[0]:
                        continue

                    try:
                        rdata = self.kaddr_space.base.read(mdata[0], mdata[2])
                    except (IOError, OverflowError):
                        debug.debug("IOError: Pid: {0} File: {1} PhysAddr: {2} Size: {3}".format(summaryinfo['pid'], summaryinfo['name'], mdata[0], mdata[2]))

                    if not rdata:
                        continue

                    of.seek(mdata[1])
                    of.write(rdata)
                    continue
                # XXX Verify FileOffsets
                #for zpad in summaryinfo['pad']:
                #    of.seek(zpad[0])
                #    of.write("\0" * zpad[1])

                if self._config.SUMMARY_FILE:
                    json.dump(summaryinfo, summaryfo)
                of.close()

            elif summaryinfo['type'] == "ImageSectionObject":
                outfd.write("ImageSectionObject {0:#010x}   {1:<6} {2}\n".format(summaryinfo['fobj'], summaryinfo['pid'], summaryinfo['name']))

                if len(summaryinfo['present']) == 0:
                    continue

                of = open(summaryinfo['ofpath'], 'wb')

                for mdata in summaryinfo['present']:
                    rdata = None
                    if not mdata[0]:
                        continue

                    try:
                        rdata = self.kaddr_space.base.read(mdata[0], mdata[2])
                    except (IOError, OverflowError):
                        debug.debug("IOError: Pid: {0} File: {1} PhysAddr: {2} Size: {3}".format(summaryinfo['pid'], summaryinfo['name'], mdata[0], mdata[2]))

                    if not rdata:
                        continue

                    of.seek(mdata[1])
                    of.write(rdata)
                    continue

                # XXX Verify FileOffsets
                #for zpad in summaryinfo['pad']:
                #    print "ZPAD 0x%x"%(zpad[0])
                #    of.seek(zpad[0])
                #    of.write("\0" * zpad[1])

                if self._config.SUMMARY_FILE:
                    json.dump(summaryinfo, summaryfo)
                of.close()

            elif summaryinfo['type'] == "SharedCacheMap":

                outfd.write("SharedCacheMap {0:#010x}   {1:<6} {2}\n".format(summaryinfo['fobj'], summaryinfo['pid'], summaryinfo['name']))
                of = open(summaryinfo['ofpath'], 'wb')
                for vacb in summaryinfo['vacbary']:
                    if not vacb:
                        continue
                    (rdata, mdata, zpad) = self.audited_read_bytes(self.kaddr_space, vacb['baseaddr'], vacb['size'], True)
                    ### We need to update the mdata,zpad
                    if rdata:
                        try:
                            of.seek(vacb['foffset'])
                            of.write(rdata)
                        except IOError:
                            # TODO: Handle things like write errors (not enough disk space, etc)
                            continue
                    vacb['present'] = mdata
                    vacb['pad'] = zpad

                if self._config.SUMMARY_FILE:
                    json.dump(summaryinfo, summaryfo)
                of.close()

            else:
                return
        if self._config.SUMMARY_FILE:
            summaryfo.close()



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
#
# Additional Authors:
# Mike Auty <mike.auty@gmail.com>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import struct
import volatility.plugins.taskmods as taskmods
import volatility.debug as debug
import volatility.obj as obj
import volatility.exceptions as exceptions

class ProcExeDump(taskmods.DllList):
    """Dump a process to an executable file sample"""
    def __init__(self, config, *args, **kwargs):
        taskmods.DllList.__init__(self, config, *args, **kwargs)
        config.add_option('DUMP-DIR', short_option = 'D', default = None,
                          cache_invalidator = False,
                          help = 'Directory in which to dump executable files')

        config.add_option("UNSAFE", short_option = "u", default = False, action = 'store_true',
                          help = 'Bypasses certain sanity checks when creating image')

    def dump_pe(self, space, base, dump_file):
        """
        Dump a PE from an AS into a file. 
        
        @param space: an AS to use
        @param base: PE base address
        @param dump_file: dumped file name

        @returns a string status message 
        """

        of = open(os.path.join(self._config.DUMP_DIR, dump_file), 'wb')
        try:
            for offset, code in self.get_image(space, base):
                of.seek(offset)
                of.write(code)
            result = "OK: {0}".format(dump_file)
        except ValueError, ve:
            result = "Error: {0}".format(ve)
        except exceptions.SanityCheckException, ve:
            result = "Error: {0} Try -u/--unsafe".format(ve)
        finally:
            of.close()

        return result

    def render_text(self, outfd, data):
        """Renders the tasks to disk images, outputting progress as they go"""
        if self._config.DUMP_DIR == None:
            debug.error("Please specify a dump directory (--dump-dir)")
        if not os.path.isdir(self._config.DUMP_DIR):
            debug.error(self._config.DUMP_DIR + " is not a directory")

        self.table_header(outfd,
                          [("Process(V)", "[addrpad]"),
                           ("ImageBase", "[addrpad]"),
                           ("Name", "20"),
                           ("Result", "")])

        for task in data:
            task_space = task.get_process_address_space()
            if task_space == None:
                result = "Error: Cannot acquire process AS"
            elif task.Peb == None:
                # we must use m() here, because any other attempt to 
                # reference task.Peb will try to instantiate the _PEB
                result = "Error: PEB at {0:#x} is paged".format(task.m('Peb'))
            elif task_space.vtop(task.Peb.ImageBaseAddress) == None:
                result = "Error: ImageBaseAddress at {0:#x} is paged".format(task.Peb.ImageBaseAddress)
            else:
                dump_file = "executable." + str(task.UniqueProcessId) + ".exe"
                result = self.dump_pe(task_space,
                                task.Peb.ImageBaseAddress,
                                dump_file)
            self.table_row(outfd,
                            task.obj_offset,
                            task.Peb.ImageBaseAddress,
                            task.ImageFileName,
                            result)

    def round(self, addr, align, up = False):
        """Rounds down an address based on an alignment"""
        if addr % align == 0:
            return addr
        else:
            if up:
                return (addr + (align - (addr % align)))
            return (addr - (addr % align))

    def get_nt_header(self, addr_space, base_addr):
        """Returns the NT Header object for a task"""

        dos_header = obj.Object("_IMAGE_DOS_HEADER", offset = base_addr,
                                vm = addr_space)

        return dos_header.get_nt_header()

    def get_code(self, addr_space, data_start, data_size, offset):
        """Returns a single section of re-created data from a file image"""
        first_block = 0x1000 - data_start % 0x1000
        full_blocks = ((data_size + (data_start % 0x1000)) / 0x1000) - 1
        left_over = (data_size + data_start) % 0x1000

        paddr = addr_space.vtop(data_start)
        code = ""

        # Deal with reads that are smaller than a block
        if data_size < first_block:
            data_read = addr_space.zread(data_start, data_size)
            if paddr == None:
                if self._config.verbose:
                    debug.debug("Memory Not Accessible: Virtual Address: 0x{0:x} File Offset: 0x{1:x} Size: 0x{2:x}\n".format(data_start, offset, data_size))
            code += data_read
            return (offset, code)

        data_read = addr_space.zread(data_start, first_block)
        if paddr == None:
            if self._config.verbose:
                debug.debug("Memory Not Accessible: Virtual Address: 0x{0:x} File Offset: 0x{1:x} Size: 0x{2:x}\n".format(data_start, offset, first_block))
        code += data_read

        # The middle part of the read
        new_vaddr = data_start + first_block

        for _i in range(0, full_blocks):
            data_read = addr_space.zread(new_vaddr, 0x1000)
            if addr_space.vtop(new_vaddr) == None:
                if self._config.verbose:
                    debug.debug("Memory Not Accessible: Virtual Address: 0x{0:x} File Offset: 0x{1:x} Size: 0x{2:x}\n".format(new_vaddr, offset, 0x1000))
            code += data_read
            new_vaddr = new_vaddr + 0x1000

        # The last part of the read
        if left_over > 0:
            data_read = addr_space.zread(new_vaddr, left_over)
            if addr_space.vtop(new_vaddr) == None:
                if self._config.verbose:
                    debug.debug("Memory Not Accessible: Virtual Address: 0x{0:x} File Offset: 0x{1:x} Size: 0x{2:x}\n".format(new_vaddr, offset, left_over))
            code += data_read
        return (offset, code)

    def get_image(self, addr_space, base_addr):
        """Outputs an executable disk image of a process"""
        nt_header = self.get_nt_header(addr_space = addr_space,
                                       base_addr = base_addr)

        soh = nt_header.OptionalHeader.SizeOfHeaders
        header = addr_space.zread(base_addr, soh)
        yield (0, header)

        fa = nt_header.OptionalHeader.FileAlignment
        for sect in nt_header.get_sections(self._config.UNSAFE):
            foa = self.round(sect.PointerToRawData, fa)
            if foa != sect.PointerToRawData:
                debug.warning("Section start on disk not aligned to file alignment.\n")
                debug.warning("Adjusted section start from {0} to {1}.\n".format(sect.PointerToRawData, foa))
            yield self.get_code(addr_space,
                                sect.VirtualAddress + base_addr,
                                sect.SizeOfRawData, foa)

class ProcMemDump(ProcExeDump):
    """Dump a process to an executable memory sample"""

    def replace_header_field(self, sect, header, item, value):
        """Replaces a field in a sector header"""
        field_size = item.size()
        start = item.obj_offset - sect.obj_offset
        end = start + field_size
        newval = struct.pack(item.format_string, int(value))
        result = header[:start] + newval + header[end:]
        return result

    def get_image(self, addr_space, base_addr):
        """Outputs an executable memory image of a process"""
        nt_header = self.get_nt_header(addr_space, base_addr)

        sa = nt_header.OptionalHeader.SectionAlignment
        shs = addr_space.profile.get_obj_size('_IMAGE_SECTION_HEADER')

        yield self.get_code(addr_space, base_addr, nt_header.OptionalHeader.SizeOfImage, 0)

        prevsect = None
        sect_sizes = []
        for sect in nt_header.get_sections(self._config.UNSAFE):
            if prevsect is not None:
                sect_sizes.append(sect.VirtualAddress - prevsect.VirtualAddress)
            prevsect = sect
        if prevsect is not None:
            sect_sizes.append(self.round(prevsect.Misc.VirtualSize, sa, up = True))

        counter = 0
        start_addr = nt_header.FileHeader.SizeOfOptionalHeader + (nt_header.OptionalHeader.obj_offset - base_addr)
        for sect in nt_header.get_sections(self._config.UNSAFE):
            sectheader = addr_space.read(sect.obj_offset, shs)
            # Change the PointerToRawData
            sectheader = self.replace_header_field(sect, sectheader, sect.PointerToRawData, sect.VirtualAddress)
            sectheader = self.replace_header_field(sect, sectheader, sect.SizeOfRawData, sect_sizes[counter])
            sectheader = self.replace_header_field(sect, sectheader, sect.Misc.VirtualSize, sect_sizes[counter])

            yield (start_addr + (counter * shs), sectheader)
            counter += 1



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization:
"""

import volatility.plugins.linux.common as linux_common
from volatility.plugins.linux.slab_info import linux_slabinfo

class linux_dentry_cache(linux_common.AbstractLinuxCommand):
    """Gather files from the dentry cache"""

    def __init__(self, config, *args, **kwargs):
        linux_common.AbstractLinuxCommand.__init__(self, config, *args, **kwargs)
        self._config.add_option('UNALLOCATED', short_option = 'u',
                        default = False,
                        help = 'Show unallocated',
                        action = 'store_true')

    def make_body(self, dentry):
        """Create a pipe-delimited bodyfile from a dentry structure. 
        
        MD5|name|inode|mode_as_string|UID|GID|size|atime|mtime|ctime|crtime
        """
        
        path = dentry.get_partial_path() or ""
        i = dentry.d_inode
        
        if i:
            ret = [0, path, i.i_ino, 0, i.i_uid, i.i_gid, i.i_size, i.i_atime, i.i_mtime, 0, i.i_ctime]
        else:
            ret = [0, path] + [0] * 8
            
        ret = "|".join([str(val) for val in ret])
        return ret

    def calculate(self):
        linux_common.set_plugin_members(self)

        cache = linux_slabinfo(self._config).get_kmem_cache("dentry", self._config.UNALLOCATED)

        # support for old kernels 
        if cache == []:
            cache = linux_slabinfo(self._config).get_kmem_cache("dentry_cache", self._config.UNALLOCATED, struct_name = "dentry")

        for dentry in cache:
            yield self.make_body(dentry)

    def render_text(self, outfd, data):

        for bodyline in data:
            outfd.write(bodyline + "\n")






# Volatility
#
# Authors:
# Mike Auty <mike.auty@gmail.com>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.conf as conf
import logging

config = conf.ConfObject()

def disable_warnings(_option, _opt_str, _value, _parser):
    """Sets the location variable in the parser to the filename in question"""
    rootlogger = logging.getLogger('')
    rootlogger.setLevel(logging.WARNING + 1)

config.add_option("WARNINGS", default = False, action = "callback",
                  callback = disable_warnings,
                  short_option = 'W', nargs = 0,
                  help = "Disable warning messages")



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import zlib
import volatility.obj as obj
import volatility.plugins.addrspaces.standard as standard

class HPAKVTypes(obj.ProfileModification):
    def modification(self, profile):
        profile.vtypes.update({
            'HPAK_HEADER' : [ 0x20, { 
                'Magic' : [ 0, ['String', dict(length = 4)]], 
                }], 
            'HPAK_SECTION': [ 0xE0, { 
                'Header' : [ 0, ['String', dict(length = 32)]], 
                'Compressed' : [ 0x8C, ['unsigned int']], 
                'Length' : [ 0x98, ['unsigned long long']], 
                'Offset' : [ 0xA8, ['unsigned long long']], 
                'NextSection' : [ 0xB0, ['unsigned long long']], 
                'Name' : [ 0xD4, ['String', dict(length = 12)]], 
                }], 
            })
        profile.object_classes.update({'HPAK_HEADER': HPAK_HEADER})
            
class HPAK_HEADER(obj.CType):
    """A class for B.S. Hairy headers"""
    
    def Sections(self):
    
        ## The initial section object 
        section = obj.Object("HPAK_SECTION", 
                             offset = self.obj_vm.profile.get_obj_size("HPAK_HEADER"), 
                             vm = self.obj_vm)
        
        ## Iterate through the sections 
        while section.is_valid():
            yield section 
            section = section.NextSection.dereference_as("HPAK_SECTION") 

class HPAKAddressSpace(standard.FileAddressSpace):
    """ This AS supports the HPAK format """
    
    order = 30
    
    def __init__(self, base, config, **kwargs):
        ## We must have an AS below us
        self.as_assert(base, "No base Address Space")
        standard.FileAddressSpace.__init__(self, base, config, layered = True, **kwargs)
                
        self.header = obj.Object("HPAK_HEADER", offset = 0, vm = base)
                
        ## Check the magic 
        self.as_assert(self.header.Magic == 'HPAK', "Invalid magic found")
                
        self.physmem = None
                
        ## cycle though looking for the PHYSDUMP header
        for section in self.header.Sections():
            if str(section.Header) == "HPAKSECTHPAK_SECTION_PHYSDUMP":
                self.physmem = section
                break
        
        self.as_assert(self.physmem is not None, "Cannot find the PHYSDUMP section")
                        
    def read(self, addr, length):
        return self.base.read(addr + self.physmem.Offset, length)
        
    def zread(self, addr, length):
        return self.base.zread(addr + self.physmem.Offset, length)
        
    def is_valid_address(self, addr):
        return self.base.is_valid_address(addr + self.physmem.Offset)
        
    def get_header(self):
        return self.header
        
    def convert_to_raw(self, outfd):
        """The standard imageinfo plugin won't work on 
        hpak images so we provide this method. It wraps
        the zlib compression if necessary"""
        
        d = zlib.decompressobj(16 + zlib.MAX_WBITS)
        
        chunk_size = 4096
        chunks = self.physmem.Length / chunk_size
        
        def get_chunk(addr, size):
            buffer = self.base.read(addr, size)
            if self.physmem.Compressed == 1:
                buffer = d.decompress(buffer)
            return buffer
        
        for i in range(chunks):
            outfd.write(get_chunk(self.physmem.Offset + i * chunk_size, chunk_size))
            yield i 
            
        leftover = self.physmem.Length % chunk_size
        
        if leftover > 0:
            outfd.write(get_chunk(self.physmem.Offset + i * chunk_size, leftover))



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.plugins.gui.windowstations as windowstations

class DeskScan(windowstations.WndScan):
    """Poolscaner for tagDESKTOP (desktops)"""

    def render_text(self, outfd, data):

        seen = []

        for window_station in data:
            for desktop in window_station.desktops():

                offset = desktop.PhysicalAddress
                if offset in seen:
                    continue
                seen.append(offset)

                outfd.write("*" * 50 + "\n")
                outfd.write("Desktop: {0:#x}, Name: {1}\\{2}, Next: {3:#x}\n".format(
                    offset,
                    desktop.WindowStation.Name,
                    desktop.Name,
                    desktop.rpdeskNext.v(),
                    ))
                outfd.write("SessionId: {0}, DesktopInfo: {1:#x}, fsHooks: {2}\n".format(
                    desktop.dwSessionId,
                    desktop.pDeskInfo.v(),
                    desktop.DeskInfo.fsHooks,
                    ))
                outfd.write("spwnd: {0:#x}, Windows: {1}\n".format(
                    desktop.DeskInfo.spwnd,
                    len(list(desktop.windows(desktop.DeskInfo.spwnd)))
                    ))
                outfd.write("Heap: {0:#x}, Size: {1:#x}, Base: {2:#x}, Limit: {3:#x}\n".format(
                    desktop.pheapDesktop.v(),
                    desktop.DeskInfo.pvDesktopLimit - desktop.DeskInfo.pvDesktopBase,
                    desktop.DeskInfo.pvDesktopBase,
                    desktop.DeskInfo.pvDesktopLimit,
                    ))
                ## This is disabled until we bring in the heaps plugin 
                #if self._config.VERBOSE:
                #    granularity = desktop.obj_vm.profile.get_obj_size("_HEAP_ENTRY")
                #    for entry in desktop.heaps():
                #        outfd.write("  Alloc: {0:#x}, Size: {1:#x} Previous: {2:#x}\n".format(
                #            entry.obj_offset + granularity, 
                #            entry.Size, entry.PreviousSize,
                #            ))
                for thrd in desktop.threads():
                    outfd.write(" {0} ({1} {2} parent {3})\n".format(
                        thrd.pEThread.Cid.UniqueThread,
                        thrd.ppi.Process.ImageFileName,
                        thrd.ppi.Process.UniqueProcessId,
                        thrd.ppi.Process.InheritedFromUniqueProcessId,
                    ))



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization: 
"""

import volatility.obj as obj
import volatility.debug as debug

import volatility.plugins.linux.common as linux_common
import volatility.plugins.linux.pslist as linux_pslist

PIDTYPE_PID = 0

# determining the processing algorithm to use is based on crash from redhat
class linux_pidhashtable(linux_pslist.linux_pslist):
    """Enumerates processes through the PID hash table"""

    def __init__(self, *args, **kwargs):
        self.seen_tasks = {}
        linux_pslist.linux_pslist.__init__(self, *args, **kwargs)

    def get_obj(self, ptr, sname, member):
        offset = self.profile.get_obj_offset(sname, member)
        addr   = ptr - offset
        return obj.Object(sname, offset = addr, vm = self.addr_space)

    def _task_for_pid(self, upid, pid):

        chained = 0

        pid_tasks_0 = pid.tasks[0].first

        if pid_tasks_0 == 0:
            chained = 1
            pnext_addr = upid.obj_offset + self.profile.get_obj_offset("upid", "pid_chain") + self.profile.get_obj_offset("hlist_node", "next")
            pnext = obj.Object("unsigned long", offset = pnext_addr, vm = self.addr_space)
            upid = obj.Object("upid", offset = pnext - self.profile.get_obj_offset("upid", "pid_chain"), vm = self.addr_space)
            for task in self._walk_upid(upid):
                yield task

        if chained == 0:
            task = obj.Object("task_struct", offset = pid_tasks_0 - self.profile.get_obj_offset("task_struct", "pids"), vm = self.addr_space)
            if task.pid > 0:
                yield task

    def _walk_upid(self, upid):

        while upid:

            pid = self.get_obj(upid.obj_offset, "pid", "numbers")

            for task in self._task_for_pid(upid, pid):
                yield task

            if type(upid.pid_chain) == obj.Pointer:
                pid_chain = obj.Object("hlist_node", offset = upid.pid_chain.obj_offset, vm = self.addr_space)
            else:
                pid_chain = upid.pid_chain

            if not pid_chain:
                break

            upid = self.get_obj(pid_chain.next, "upid", "pid_chain")

    def calculate_v3(self):
        self.seen_tasks = {}

        pidhash_shift = obj.Object("unsigned int", offset = self.addr_space.profile.get_symbol("pidhash_shift"), vm = self.addr_space)
        pidhash_size = 1 << pidhash_shift

        pidhash_addr = self.addr_space.profile.get_symbol("pid_hash")
        pidhash_ptr = obj.Object("Pointer", offset = pidhash_addr, vm = self.addr_space)

        # pidhash is an array of hlist_heads
        pidhash = obj.Object(theType = 'Array', offset = pidhash_ptr, vm = self.addr_space, targetType = 'hlist_head', count = pidhash_size)

        for hlist in pidhash:

            # each entry in the hlist is a upid which is wrapped in a pid
            ent = hlist.first

            while ent.v():
                upid = self.get_obj(ent.obj_offset, "upid", "pid_chain")

                for task in self._walk_upid(upid):
                    if not task.obj_offset in self.seen_tasks:
                        self.seen_tasks[task.obj_offset] = 1
                        if task.is_valid_task():
                            yield task

                ent = ent.m("next")

    # the following functions exist because crash has handlers for them
    # but I was unable to find a profile/kernel that needed them (maybe too old or just a one-off distro kernel
    # if someone actually triggers this message, I can quickly add in the support as I will have a sample to test again
    def profile_unsupported(self, func_name):
        debug.error("{0:s}: This profile is currently unsupported by this plugin. Please file a bug report on our issue tracker to have support added.".format(func_name))

    def calculate_v2(self):
        self.profile_unsupported("calculate_v2")

    def calculate_v1(self):
        self.profile_unsupported("calculate_v1")

    def refresh_pid_hash_task_table(self):
        self.profile_unsupported("refresh_pid_hash_task_table")

    def get_both(self):

        has_pid_link = self.profile.has_type("pid_link")
        has_link_pid = self.profile.obj_has_member("pid_link", "pid")

        has_pid_hash = self.profile.has_type("pid_hash")
        has_upid = self.profile.has_type("upid")
        has_pid_numbers = self.profile.obj_has_member("pid", "numbers")

        if has_pid_hash:
            has_hash_chain = self.profile.obj_has_member("pid_hash", "chain")
        else:
            has_hash_chain = None

        if has_link_pid and has_hash_chain:
            func = self.refresh_pid_hash_task_table

        elif has_pid_link:
            if has_upid and has_pid_numbers:
                func = self.calculate_v3 # refresh_hlist_task_table_v3
            else:
                func = self.calculate_v2 # refresh_hlist_task_table_v2
        else:
            func = self.calculate_v1

        return func

    def determine_func(self):

        pidhash = self.addr_space.profile.get_symbol("pidhash")
        pid_hash = self.addr_space.profile.get_symbol("pid_hash")
        pidhash_shift = self.addr_space.profile.get_symbol("pidhash_shift")

        if pid_hash and pidhash_shift:
            func = self.get_both()

        elif pid_hash:
            func = self.refresh_pid_hash_task_table

        elif pidhash:
            func = self.refresh_pid_hash_task_table

        return func

    def calculate(self):
        linux_common.set_plugin_members(self)
        func = self.determine_func()

        for task in func():
            yield task









# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.plugins.malware.malfind as malfind
import volatility.plugins.mac.pstasks as pstasks
import volatility.plugins.mac.common as common
import volatility.utils as utils 
import volatility.debug as debug
import volatility.obj as obj

try:
    import yara
    has_yara = True
except ImportError:
    has_yara = False

class MapYaraScanner(malfind.BaseYaraScanner):
    """A scanner over all memory regions of a process."""

    def __init__(self, task = None, **kwargs):
        """Scan the process address space through the VMAs.

        Args:
          task: The task_struct object for this task.
        """
        self.task = task
        malfind.BaseYaraScanner.__init__(self, address_space = task.get_process_address_space(), **kwargs)

    def scan(self, offset = 0, maxlen = None):
        for map in self.task.get_proc_maps():
            for match in malfind.BaseYaraScanner.scan(self, map.links.start, map.links.end - map.links.start):
                yield match

class mac_yarascan(malfind.YaraScan):
    """Scan memory for yara signatures"""

    @staticmethod
    def is_valid_profile(profile):
        return profile.metadata.get('os', 'Unknown').lower() == 'mac'

    def calculate(self):
    
        ## we need this module imported
        if not has_yara:
            debug.error("Please install Yara from code.google.com/p/yara-project")
            
        ## leveraged from the windows yarascan plugin
        rules = self._compile_rules()
            
        ## set the linux plugin address spaces 
        common.set_plugin_members(self)

        if self._config.KERNEL:
            ## http://fxr.watson.org/fxr/source/osfmk/mach/i386/vm_param.h?v=xnu-2050.18.24
            if self.addr_space.profile.metadata.get('memory_model', '32bit') == "32bit":
                if not common.is_64bit_capable(self.addr_space):
                    kernel_start = 0
                else:
                    kernel_start = 0xc0000000
            else:
                kernel_start = 0xffffff8000000000

            scanner = malfind.DiscontigYaraScanner(rules = rules, 
                                                   address_space = self.addr_space) 
      
            for hit, address in scanner.scan(start_offset = kernel_start):
                yield (None, address, hit, 
                        scanner.address_space.zread(address, 64))
        else:
            # Scan each process memory block 
            for task in pstasks.mac_tasks(self._config).calculate():
                scanner = MapYaraScanner(task = task, rules = rules)
                for hit, address in scanner.scan():
                    yield (task, address, hit, 
                            scanner.address_space.zread(address, 64))
    
    def render_text(self, outfd, data):
        for task, address, hit, buf in data:
            if task:
                outfd.write("Task: {0} pid {1} rule {2} addr {3:#x}\n".format(
                    task.p_comm, task.p_pid, hit.rule, address))
            else:
                outfd.write("[kernel] rule {0} addr {1:#x}\n".format(hit.rule, address))
            
            outfd.write("".join(["{0:#018x}  {1:<48}  {2}\n".format(
                address + o, h, ''.join(c)) for o, h, c in utils.Hexdump(buf)]))



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization: 
"""

import volatility.obj as obj
import volatility.plugins.linux.common as linux_common

class linux_check_idt(linux_common.AbstractLinuxCommand):
    """ Checks if the IDT has been altered """

    def calculate(self):
        """ 
        This works by walking the IDT table for the entries that Linux uses
        and verifies that each is a symbol in the kernel
        """
        linux_common.set_plugin_members(self)

        tblsz = 256

        sym_addrs = self.profile.get_all_addresses()

        # hw handlers + system call
        check_idxs = list(range(0, 20)) + [128]

        if self.profile.metadata.get('memory_model', '32bit') == "32bit":
            idt_type = "desc_struct"
        else:
            idt_type = "gate_struct64"

        # this is written as a list b/c there are supposdly kernels with per-CPU IDTs
        # but I haven't found one yet...
        addrs = [self.addr_space.profile.get_symbol("idt_table")]

        for tableaddr in addrs:

            table = obj.Object(theType = 'Array', offset = tableaddr, vm = self.addr_space, targetType = idt_type, count = tblsz)

            for i in check_idxs:

                ent = table[i]

                if not ent:
                    continue

                idt_addr = ent.Address

                if idt_addr != 0:
                    if not idt_addr in sym_addrs:
                        hooked = 1
                        sym_name = "HOOKED"
                    else:
                        hooked = 0
                        sym_name = self.profile.get_symbol_by_address("kernel", idt_addr)

                    yield(i, idt_addr, sym_name, hooked)

    def render_text(self, outfd, data):

        self.table_header(outfd, [("Index", "[addr]"), ("Address", "[addrpad]"), ("Symbol", "<30")])

        for (i, idt_addr, sym_name, hooked) in data:
            self.table_row(outfd, i, idt_addr, sym_name)







# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.obj as obj

class Win2003x86GuiVTypes(obj.ProfileModification):
    """Apply the overlays for Windows 2003 x86 (builds on Windows XP x86)"""

    before = ["XP2003x86BaseVTypes"]

    conditions = {'os': lambda x: x == 'windows',
                  'memory_model': lambda x: x == '32bit',
                  'major': lambda x: x == 5,
                  'minor': lambda x: x == 2}

    def modification(self, profile):

        profile.merge_overlay({
            'tagWINDOWSTATION' : [ 0x54, {
            'spwndClipOwner' : [ 0x18, ['pointer', ['tagWND']]],
            'pGlobalAtomTable' : [ 0x3C, ['pointer', ['void']]],
            }],
            'tagTHREADINFO' : [ None, {
            'PtiLink' : [ 0xB0, ['_LIST_ENTRY']],
            'fsHooks' : [ 0x9C, ['unsigned long']],
            'aphkStart' : [ 0xF8, ['array', 16, ['pointer', ['tagHOOK']]]],
            }],
            'tagDESKTOP' : [ None, {
            'hsectionDesktop' : [ 0x3c, ['pointer', ['void']]],
            'pheapDesktop' : [ 0x40, ['pointer', ['tagWIN32HEAP']]],
            'ulHeapSize' : [ 0x44, ['unsigned long']],
            'PtiList' : [ 0x60, ['_LIST_ENTRY']],
            }],
            'tagSERVERINFO' : [ None, {
            'cHandleEntries' : [ 4, ['unsigned long']],
            'cbHandleTable' : [ 0x1b8, ['unsigned long']],
            }],
        })




# Volatility
# Copyright (c) 2008-2013 Volatility Foundation
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0
@contact:      bdolangavitt@wesleyan.edu

This file provides support for Windows XP.
"""

#pylint: disable-msg=C0111

import volatility.debug as debug #pylint: disable-msg=W0611
import volatility.obj as obj

class XPOverlay(obj.ProfileModification):
    before = ['WindowsOverlay']
    conditions = {'os': lambda x : x == 'windows',
                  'major': lambda x: x == 5,
                  'minor': lambda x: x == 1}

    def modification(self, profile):
        overlay = {'VOLATILITY_MAGIC': [ None, {
                        'DTBSignature' : [ None, ['VolatilityMagic', dict(value = "\x03\x00\x1b\x00")]],
                        'KDBGHeader'   : [ None, ['VolatilityMagic', dict(value = '\x00\x00\x00\x00\x00\x00\x00\x00KDBG\x90\x02')]],
                        'HibrProcPage' : [ None, ['VolatilityMagic', dict(value = 0x2)]],
                        'HibrEntryCount' : [ None, ['VolatilityMagic', dict(value = 0xff)]],
                                                }],
                   '_EPROCESS'        : [ None, {
                        'VadRoot'      : [ None, ['pointer', ['_MMVAD']]]
                                                }]
                      }
        profile.merge_overlay(overlay)

class WinXPSP2x86(obj.Profile):
    """ A Profile for Windows XP SP2 x86 """
    _md_major = 5
    _md_minor = 1
    _md_os = 'windows'
    _md_memory_model = '32bit'
    _md_vtype_module = 'volatility.plugins.overlays.windows.xp_sp2_x86_vtypes'

class WinXPSP3x86(obj.Profile):
    """ A Profile for Windows XP SP3 x86 """
    _md_major = 5
    _md_minor = 1
    _md_os = 'windows'
    _md_memory_model = '32bit'
    _md_vtype_module = 'volatility.plugins.overlays.windows.xp_sp3_x86_vtypes'





# Volatility
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.obj as obj

class _KDDEBUGGER_DATA64(obj.CType):
    """A class for KDBG"""

    def is_valid(self):
        """Returns true if the kdbg_object appears valid"""
        # Check the OwnerTag is in fact the string KDBG
        return obj.CType.is_valid(self) and self.Header.OwnerTag == 0x4742444B

    @property
    def ServicePack(self):
        """Get the service pack number. This is something
        like 0x100 for SP1, 0x200 for SP2 etc. 
        """
        csdresult = obj.Object("unsigned long", offset = self.CmNtCSDVersion, vm = self.obj_vm)
        return (csdresult >> 8) & 0xffffffff

    def processes(self):
        """Enumerate processes"""
        # This is defined as a pointer to _LIST_ENTRY in the overlay
        list_head = self.PsActiveProcessHead.dereference()
        if not list_head:
            raise AttributeError("Could not list tasks, please verify your --profile with kdbgscan")

        for l in list_head.list_of_type("_EPROCESS", "ActiveProcessLinks"):
            yield l

    def modules(self):
        """Enumerate modules"""
        # This is defined as a pointer to _LIST_ENTRY in the overlay
        list_head = self.PsLoadedModuleList.dereference()
        if not list_head:
            raise AttributeError("Could not list modules, please verify your --profile with kdbgscan")

        for l in list_head.dereference_as("_LIST_ENTRY").list_of_type(
            "_LDR_DATA_TABLE_ENTRY", "InLoadOrderLinks"):
            yield l

    def dbgkd_version64(self):
        """Scan backwards from the base of KDBG to find the 
        _DBGKD_GET_VERSION64. We have a winner when kernel 
        base addresses and process list head match."""

        # Account for address masking differences in x86 and x64
        memory_model = self.obj_vm.profile.metadata.get('memory_model', '32bit')

        dbgkd_off = self.obj_offset & 0xFFFFFFFFFFFFF000
        dbgkd_end = dbgkd_off + 0x1000
        # The _DBGKD_GET_VERSION64 structure is autogenerated, so
        # this value should be correct for each profile
        dbgkd_size = self.obj_vm.profile.get_obj_size("_DBGKD_GET_VERSION64")

        while dbgkd_off <= (dbgkd_end - dbgkd_size):

            dbgkd = obj.Object("_DBGKD_GET_VERSION64",
                        offset = dbgkd_off,
                        vm = self.obj_vm)

            if memory_model == "32bit":
                KernBase = dbgkd.KernBase & 0xFFFFFFFF
                PsLoadedModuleList = dbgkd.PsLoadedModuleList & 0xFFFFFFFF
            else:
                KernBase = dbgkd.KernBase
                PsLoadedModuleList = dbgkd.PsLoadedModuleList

            if ((KernBase == self.KernBase) and (PsLoadedModuleList == self.PsLoadedModuleList)):
                return dbgkd

            dbgkd_off += 1

        return obj.NoneObject("Cannot find _DBGKD_GET_VERSION64")

    def kpcrs(self):
        """Generator for KPCRs referenced by this KDBG. 

        These are returned in the order in which the 
        processors were registered. 
        """

        if self.obj_vm.profile.metadata.get('memory_model', '32bit') == '32bit':
            prcb_member = "PrcbData"
        else:
            prcb_member = "Prcb"

        cpu_array = self.KiProcessorBlock.dereference()

        for p in cpu_array:

            # Terminate the loop if an item in the array is 
            # invalid (ie paged) or if the pointer is NULL. 
            if p == None or p == 0:
                break

            kpcrb = p.dereference_as("_KPRCB")

            yield obj.Object("_KPCR", offset = kpcrb.obj_offset -
                    self.obj_vm.profile.get_obj_offset("_KPCR", prcb_member),
                    vm = self.obj_vm,
                    )

class KDBGObjectClass(obj.ProfileModification):
    """Add the KDBG object class to all Windows profiles"""

    before = ["WindowsObjectClasses"]
    conditions = {'os': lambda x: x == 'windows'}

    def modification(self, profile):
        profile.object_classes.update({'_KDDEBUGGER_DATA64': _KDDEBUGGER_DATA64})

        # This value is stored in nt!_KeMaximumProcessors
        if profile.metadata.get('memory_model', '32bit'):
            max_processors = 32
        else:
            max_processors = 64

        profile.merge_overlay({
            '_KDDEBUGGER_DATA64': [ None, {
            'NtBuildLab': [ None, ['pointer', ['String', dict(length = 32)]]],
            'KiProcessorBlock': [ None, ['pointer', ['array', max_processors, ['pointer', ['_KPRCB']]]]],
            'PsActiveProcessHead': [ None, ['pointer', ['_LIST_ENTRY']]], 
            'PsLoadedModuleList': [ None, ['pointer', ['_LIST_ENTRY']]], 
            'MmUnloadedDrivers' : [ None, ['pointer', ['pointer', ['array', lambda x : x.MmLastUnloadedDriver.dereference(), ['_UNLOADED_DRIVER']]]]], 
            'MmLastUnloadedDriver' : [ None, ['pointer', ['unsigned int']]], 
            }]})

class UnloadedDriverVTypes(obj.ProfileModification):
    """Add the unloaded driver structure definitions"""

    conditions = {'os': lambda x: x == "windows"} 

    def modification(self, profile):

        if profile.metadata.get("memory_model", "32bit") == "32bit":
            vtypes = {'_UNLOADED_DRIVER' : [ 24, { 
                    'Name' : [ 0, ['_UNICODE_STRING']], 
                    'StartAddress' : [ 8, ['address']], 
                    'EndAddress' : [ 12, ['address']], 
                    'CurrentTime' : [ 16, ['WinTimeStamp', {}]], 
                    }]}
        else:
            vtypes = {'_UNLOADED_DRIVER' : [ 40, { 
                    'Name' : [ 0, ['_UNICODE_STRING']], 
                    'StartAddress' : [ 16, ['address']], 
                    'EndAddress' : [ 24, ['address']], 
                    'CurrentTime' : [ 32, ['WinTimeStamp', {}]], 
                    }]}

        profile.vtypes.update(vtypes)

kdbg_vtypes = {
'_DBGKD_DEBUG_DATA_HEADER64' : [  0x18, {
  'List' : [ 0x0, ['LIST_ENTRY64']],
  'OwnerTag' : [ 0x10, ['unsigned long']],
  'Size' : [ 0x14, ['unsigned long']],
} ],
'_KDDEBUGGER_DATA64' : [  0x340, {
  'Header' : [ 0x0, ['_DBGKD_DEBUG_DATA_HEADER64']],
  'KernBase' : [ 0x18, ['unsigned long long']],
  'BreakpointWithStatus' : [ 0x20, ['unsigned long long']],
  'SavedContext' : [ 0x28, ['unsigned long long']],
  'ThCallbackStack' : [ 0x30, ['unsigned short']],
  'NextCallback' : [ 0x32, ['unsigned short']],
  'FramePointer' : [ 0x34, ['unsigned short']],
  'KiCallUserMode' : [ 0x38, ['unsigned long long']],
  'KeUserCallbackDispatcher' : [ 0x40, ['unsigned long long']],
  'PsLoadedModuleList' : [ 0x48, ['unsigned long long']],
  'PsActiveProcessHead' : [ 0x50, ['unsigned long long']],
  'PspCidTable' : [ 0x58, ['unsigned long long']],
  'ExpSystemResourcesList' : [ 0x60, ['unsigned long long']],
  'ExpPagedPoolDescriptor' : [ 0x68, ['unsigned long long']],
  'ExpNumberOfPagedPools' : [ 0x70, ['unsigned long long']],
  'KeTimeIncrement' : [ 0x78, ['unsigned long long']],
  'KeBugCheckCallbackListHead' : [ 0x80, ['unsigned long long']],
  'KiBugcheckData' : [ 0x88, ['unsigned long long']],
  'IopErrorLogListHead' : [ 0x90, ['unsigned long long']],
  'ObpRootDirectoryObject' : [ 0x98, ['unsigned long long']],
  'ObpTypeObjectType' : [ 0xa0, ['unsigned long long']],
  'MmSystemCacheStart' : [ 0xa8, ['unsigned long long']],
  'MmSystemCacheEnd' : [ 0xb0, ['unsigned long long']],
  'MmSystemCacheWs' : [ 0xb8, ['unsigned long long']],
  'MmPfnDatabase' : [ 0xc0, ['unsigned long long']],
  'MmSystemPtesStart' : [ 0xc8, ['unsigned long long']],
  'MmSystemPtesEnd' : [ 0xd0, ['unsigned long long']],
  'MmSubsectionBase' : [ 0xd8, ['unsigned long long']],
  'MmNumberOfPagingFiles' : [ 0xe0, ['unsigned long long']],
  'MmLowestPhysicalPage' : [ 0xe8, ['unsigned long long']],
  'MmHighestPhysicalPage' : [ 0xf0, ['unsigned long long']],
  'MmNumberOfPhysicalPages' : [ 0xf8, ['unsigned long long']],
  'MmMaximumNonPagedPoolInBytes' : [ 0x100, ['unsigned long long']],
  'MmNonPagedSystemStart' : [ 0x108, ['unsigned long long']],
  'MmNonPagedPoolStart' : [ 0x110, ['unsigned long long']],
  'MmNonPagedPoolEnd' : [ 0x118, ['unsigned long long']],
  'MmPagedPoolStart' : [ 0x120, ['unsigned long long']],
  'MmPagedPoolEnd' : [ 0x128, ['unsigned long long']],
  'MmPagedPoolInformation' : [ 0x130, ['unsigned long long']],
  'MmPageSize' : [ 0x138, ['unsigned long long']],
  'MmSizeOfPagedPoolInBytes' : [ 0x140, ['unsigned long long']],
  'MmTotalCommitLimit' : [ 0x148, ['unsigned long long']],
  'MmTotalCommittedPages' : [ 0x150, ['unsigned long long']],
  'MmSharedCommit' : [ 0x158, ['unsigned long long']],
  'MmDriverCommit' : [ 0x160, ['unsigned long long']],
  'MmProcessCommit' : [ 0x168, ['unsigned long long']],
  'MmPagedPoolCommit' : [ 0x170, ['unsigned long long']],
  'MmExtendedCommit' : [ 0x178, ['unsigned long long']],
  'MmZeroedPageListHead' : [ 0x180, ['unsigned long long']],
  'MmFreePageListHead' : [ 0x188, ['unsigned long long']],
  'MmStandbyPageListHead' : [ 0x190, ['unsigned long long']],
  'MmModifiedPageListHead' : [ 0x198, ['unsigned long long']],
  'MmModifiedNoWritePageListHead' : [ 0x1a0, ['unsigned long long']],
  'MmAvailablePages' : [ 0x1a8, ['unsigned long long']],
  'MmResidentAvailablePages' : [ 0x1b0, ['unsigned long long']],
  'PoolTrackTable' : [ 0x1b8, ['unsigned long long']],
  'NonPagedPoolDescriptor' : [ 0x1c0, ['unsigned long long']],
  'MmHighestUserAddress' : [ 0x1c8, ['unsigned long long']],
  'MmSystemRangeStart' : [ 0x1d0, ['unsigned long long']],
  'MmUserProbeAddress' : [ 0x1d8, ['unsigned long long']],
  'KdPrintCircularBuffer' : [ 0x1e0, ['unsigned long long']],
  'KdPrintCircularBufferEnd' : [ 0x1e8, ['unsigned long long']],
  'KdPrintWritePointer' : [ 0x1f0, ['unsigned long long']],
  'KdPrintRolloverCount' : [ 0x1f8, ['unsigned long long']],
  'MmLoadedUserImageList' : [ 0x200, ['unsigned long long']],
  'NtBuildLab' : [ 0x208, ['unsigned long long']],
  'KiNormalSystemCall' : [ 0x210, ['unsigned long long']],
  'KiProcessorBlock' : [ 0x218, ['unsigned long long']],
  'MmUnloadedDrivers' : [ 0x220, ['unsigned long long']],
  'MmLastUnloadedDriver' : [ 0x228, ['unsigned long long']],
  'MmTriageActionTaken' : [ 0x230, ['unsigned long long']],
  'MmSpecialPoolTag' : [ 0x238, ['unsigned long long']],
  'KernelVerifier' : [ 0x240, ['unsigned long long']],
  'MmVerifierData' : [ 0x248, ['unsigned long long']],
  'MmAllocatedNonPagedPool' : [ 0x250, ['unsigned long long']],
  'MmPeakCommitment' : [ 0x258, ['unsigned long long']],
  'MmTotalCommitLimitMaximum' : [ 0x260, ['unsigned long long']],
  'CmNtCSDVersion' : [ 0x268, ['unsigned long long']],
  'MmPhysicalMemoryBlock' : [ 0x270, ['unsigned long long']],
  'MmSessionBase' : [ 0x278, ['unsigned long long']],
  'MmSessionSize' : [ 0x280, ['unsigned long long']],
  'MmSystemParentTablePage' : [ 0x288, ['unsigned long long']],
  'MmVirtualTranslationBase' : [ 0x290, ['unsigned long long']],
  'OffsetKThreadNextProcessor' : [ 0x298, ['unsigned short']],
  'OffsetKThreadTeb' : [ 0x29a, ['unsigned short']],
  'OffsetKThreadKernelStack' : [ 0x29c, ['unsigned short']],
  'OffsetKThreadInitialStack' : [ 0x29e, ['unsigned short']],
  'OffsetKThreadApcProcess' : [ 0x2a0, ['unsigned short']],
  'OffsetKThreadState' : [ 0x2a2, ['unsigned short']],
  'OffsetKThreadBStore' : [ 0x2a4, ['unsigned short']],
  'OffsetKThreadBStoreLimit' : [ 0x2a6, ['unsigned short']],
  'SizeEProcess' : [ 0x2a8, ['unsigned short']],
  'OffsetEprocessPeb' : [ 0x2aa, ['unsigned short']],
  'OffsetEprocessParentCID' : [ 0x2ac, ['unsigned short']],
  'OffsetEprocessDirectoryTableBase' : [ 0x2ae, ['unsigned short']],
  'SizePrcb' : [ 0x2b0, ['unsigned short']],
  'OffsetPrcbDpcRoutine' : [ 0x2b2, ['unsigned short']],
  'OffsetPrcbCurrentThread' : [ 0x2b4, ['unsigned short']],
  'OffsetPrcbMhz' : [ 0x2b6, ['unsigned short']],
  'OffsetPrcbCpuType' : [ 0x2b8, ['unsigned short']],
  'OffsetPrcbVendorString' : [ 0x2ba, ['unsigned short']],
  'OffsetPrcbProcStateContext' : [ 0x2bc, ['unsigned short']],
  'OffsetPrcbNumber' : [ 0x2be, ['unsigned short']],
  'SizeEThread' : [ 0x2c0, ['unsigned short']],
  'KdPrintCircularBufferPtr' : [ 0x2c8, ['unsigned long long']],
  'KdPrintBufferSize' : [ 0x2d0, ['unsigned long long']],
  'KeLoaderBlock' : [ 0x2d8, ['unsigned long long']],
  'SizePcr' : [ 0x2e0, ['unsigned short']],
  'OffsetPcrSelfPcr' : [ 0x2e2, ['unsigned short']],
  'OffsetPcrCurrentPrcb' : [ 0x2e4, ['unsigned short']],
  'OffsetPcrContainedPrcb' : [ 0x2e6, ['unsigned short']],
  'OffsetPcrInitialBStore' : [ 0x2e8, ['unsigned short']],
  'OffsetPcrBStoreLimit' : [ 0x2ea, ['unsigned short']],
  'OffsetPcrInitialStack' : [ 0x2ec, ['unsigned short']],
  'OffsetPcrStackLimit' : [ 0x2ee, ['unsigned short']],
  'OffsetPrcbPcrPage' : [ 0x2f0, ['unsigned short']],
  'OffsetPrcbProcStateSpecialReg' : [ 0x2f2, ['unsigned short']],
  'GdtR0Code' : [ 0x2f4, ['unsigned short']],
  'GdtR0Data' : [ 0x2f6, ['unsigned short']],
  'GdtR0Pcr' : [ 0x2f8, ['unsigned short']],
  'GdtR3Code' : [ 0x2fa, ['unsigned short']],
  'GdtR3Data' : [ 0x2fc, ['unsigned short']],
  'GdtR3Teb' : [ 0x2fe, ['unsigned short']],
  'GdtLdt' : [ 0x300, ['unsigned short']],
  'GdtTss' : [ 0x302, ['unsigned short']],
  'Gdt64R3CmCode' : [ 0x304, ['unsigned short']],
  'Gdt64R3CmTeb' : [ 0x306, ['unsigned short']],
  'IopNumTriageDumpDataBlocks' : [ 0x308, ['unsigned long long']],
  'IopTriageDumpDataBlocks' : [ 0x310, ['unsigned long long']],
  'VfCrashDataBlock' : [ 0x318, ['unsigned long long']],
  'MmBadPagesDetected' : [ 0x320, ['unsigned long long']],
  'MmZeroedPageSingleBitErrorsDetected' : [ 0x328, ['unsigned long long']],
  'EtwpDebuggerData' : [ 0x330, ['unsigned long long']],
  'OffsetPrcbContext' : [ 0x338, ['unsigned short']],
} ],
}
